"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import logging

import numba as nb
import numpy as np
import pytest

from pde import FieldCollection, ScalarField, Tensor2Field, UnitGrid, VectorField
from pde.tools.expressions import (
    BCDataError,
    ScalarExpression,
    TensorExpression,
    evaluate,
    parse_expr_guarded,
    parse_number,
)


def test_parse_number():
    """test parse_number function"""
    assert parse_number(0) == pytest.approx(0)
    assert parse_number(1.235) == pytest.approx(1.235)
    assert parse_number("0") == pytest.approx(0)
    assert parse_number("1.235") == pytest.approx(1.235)
    assert parse_number("-1.4e-5") == pytest.approx(-1.4e-5)
    assert parse_number("sin(3.4)") == pytest.approx(np.sin(3.4))
    assert parse_number("sin(a)", {"a": 3}) == pytest.approx(np.sin(3))
    assert parse_number("a**b", {"a": 3.2, "b": 4.5}) == pytest.approx(3.2**4.5)
    assert parse_number("1 + 2 * I") == pytest.approx(1 + 2j)
    assert parse_number("a", {"a": 1 + 2j}) == pytest.approx(1 + 2j)
    with pytest.raises(TypeError):
        parse_number("foo")


def test_parse_expr_guarded():
    """test parse_expr_guarded function"""
    peg = parse_expr_guarded
    assert peg("1") == 1
    assert peg("1 + 1") == 2
    assert peg("a + 1").subs("a", 1) == 2
    with pytest.raises(TypeError):
        peg("field + 1")
    assert peg("field + 1", symbols=["field"]).subs("field", 1) == 2
    assert peg("field + 1", symbols=[None, ["field"]]).subs("field", 1) == 2
    assert peg("field + 1", symbols=["field", "a"]).subs("field", 1) == 2
    assert peg("field + 1", symbols={"field": 1, "a": 2}).subs("field", 1) == 2


def test_const(caplog):
    """test simple expressions"""
    # test scalar expressions with constants
    for expr in [None, 1, "1", "a - a"]:
        e = ScalarExpression() if expr is None else ScalarExpression(expr)
        val = 0 if expr is None or expr == "a - a" else float(expr)
        assert e.constant
        assert e.value == val
        assert e() == val
        assert e.get_compiled()() == val
        assert not e.depends_on("a")
        assert e.differentiate("a").value == 0
        assert e.shape == tuple()
        assert e.rank == 0
        assert bool(e) == (val != 0)
        assert e.is_zero == (val == 0)
        assert not e.complex

        g = e.derivatives
        assert g.constant
        assert isinstance(str(g), str)
        np.testing.assert_equal(g.value, [])

        for f in [
            ScalarExpression(e),
            ScalarExpression(e.expression),
            ScalarExpression(e.value),
        ]:
            assert e is not f
            assert e._sympy_expr == f._sympy_expr

    # test whether wrong constants are check for
    field = ScalarField(UnitGrid([3]))
    e = ScalarExpression("scalar_field", consts={"scalar_field": field})
    with caplog.at_level(logging.WARNING):
        assert e() == field
    assert "field" in caplog.text
    if not nb.config.DISABLE_JIT:
        with pytest.raises(Exception):
            e.get_compiled()()


def test_single_arg(rng):
    """test simple expressions"""
    e = ScalarExpression("2 * a")
    assert not e.constant
    assert e.depends_on("a")
    assert e(4) == 8
    assert e.get_compiled()(4) == 8
    assert e.differentiate("a").value == 2
    assert e.differentiate("b").value == 0
    assert e.shape == tuple()
    assert e.rank == 0
    assert bool(e)
    assert not e.is_zero

    assert e == ScalarExpression(e.expression)
    with pytest.raises(TypeError):
        e.value

    arr = rng.random(5)
    np.testing.assert_allclose(e(arr), 2 * arr)
    np.testing.assert_allclose(e.get_compiled()(arr), 2 * arr)

    g = e.derivatives
    assert g.shape == (1,)
    assert g.constant
    assert g(3) == [2]
    assert g.get_compiled()(3) == [2]

    with pytest.raises(TypeError):
        ScalarExpression(np.exp)


def test_two_args(rng):
    """test simple expressions"""
    e = ScalarExpression("2 * a ** b")
    assert e.depends_on("b")
    assert not e.constant
    assert e(4, 2) == 32
    assert e.get_compiled()(4, 2) == 32
    assert e.differentiate("a")(4, 2) == 16
    assert e.differentiate("b")(4, 2) == pytest.approx(32 * np.log(4))
    assert e.differentiate("c").value == 0
    assert e.shape == tuple()
    assert e.rank == 0
    assert e == ScalarExpression(e.expression)

    for x in [rng.random(2), rng.random((2, 5))]:
        res = 2 * x[0] ** x[1]
        np.testing.assert_allclose(e(*x), res)
        np.testing.assert_allclose(e.get_compiled()(*x), res)
        if x.ndim == 1:
            func = e._get_function(single_arg=True)
            np.testing.assert_allclose(func(x), res)
            func = e.get_compiled(single_arg=True)
            np.testing.assert_allclose(func(x), res)

    g = e.derivatives
    assert g.shape == (2,)
    assert g.rank == 1
    assert not g.constant
    np.testing.assert_allclose(g(2, 3), [24, 16 * np.log(2)])
    np.testing.assert_allclose(g.get_compiled()(2, 3), [24, 16 * np.log(2)])


def test_derivatives():
    """test vector expressions"""
    e = ScalarExpression("a * b**2")
    assert e.depends_on("a") and e.depends_on("b")
    assert not e.constant
    assert e.rank == 0

    d1 = e.derivatives
    assert d1.shape == (2,)
    np.testing.assert_allclose(d1(2, 3), [9, 12])
    np.testing.assert_allclose(d1.get_compiled()(2, 3), [9, 12])

    d2 = d1.derivatives
    assert d2.shape == (2, 2)
    np.testing.assert_allclose(d2(2, 3), [[0, 6], [6, 4]])
    np.testing.assert_allclose(d2.get_compiled()(2, 3), [[0, 6], [6, 4]])

    d3 = d2.derivatives
    assert d3.shape == (2, 2, 2)

    d4 = d3.derivatives
    assert d4.shape == (2, 2, 2, 2)
    np.testing.assert_allclose(d4(2, 3), np.zeros((2, 2, 2, 2)))
    np.testing.assert_allclose(d4.get_compiled()(2, 3), np.zeros((2, 2, 2, 2)))


def test_indexed():
    """test simple expressions"""
    e = ScalarExpression("2 * a[0] ** a[1]", allow_indexed=True)
    assert not e.constant
    assert e.depends_on("a")

    a = np.array([4, 2])
    assert e(a) == 32
    assert e.get_compiled()(a) == 32

    assert e.differentiate("a[0]")(a) == 16
    assert e.differentiate("a[1]")(a) == pytest.approx(32 * np.log(4))

    with pytest.raises(RuntimeError):
        e.differentiate("a")
    with pytest.raises(RuntimeError):
        e.derivatives


def test_synonyms(caplog):
    """test using synonyms in expression"""
    e = ScalarExpression("2 * arbitrary", [["a", "arbitrary"]])
    assert e.depends_on("a")
    assert not e.depends_on("arbitrary")


def test_tensor_expression():
    """test TensorExpression"""
    e = TensorExpression("[[0, 1], [2, 3]]")
    assert isinstance(str(e), str)
    assert e.shape == (2, 2)
    assert e.rank == 2
    assert e.constant
    np.testing.assert_allclose(e.get_compiled_array()(), [[0, 1], [2, 3]])
    np.testing.assert_allclose(e.get_compiled_array()(tuple()), [[0, 1], [2, 3]])
    assert e.differentiate("a") == TensorExpression("[[0, 0], [0, 0]]")
    np.testing.assert_allclose(e.value, np.arange(4).reshape(2, 2))

    e = TensorExpression("[a, 2*a]")
    assert isinstance(str(e), str)
    assert e.shape == (2,)
    assert e.rank == 1
    assert e.depends_on("a")
    assert not e.constant
    np.testing.assert_allclose(e.differentiate("a").value, np.array([1, 2]))
    with pytest.raises(TypeError):
        e.value
    assert e[0] == ScalarExpression("a")
    assert e[1] == ScalarExpression("2*a")
    assert e[0:1] == TensorExpression("[a]")
    np.testing.assert_allclose(e.get_compiled_array()(1.0), [1.0, 2.0])
    np.testing.assert_allclose(e.get_compiled_array()(2.0), [2.0, 4.0])

    e2 = TensorExpression(e)
    assert isinstance(str(e2), str)
    assert e == e2
    assert e is not e2


def test_expression_from_expression():
    """test creating expressions from expressions"""
    expr = ScalarExpression("sin(a)")
    assert expr == ScalarExpression(expr)
    assert expr != ScalarExpression(expr, ["a", "b"])
    with pytest.raises(RuntimeError):
        ScalarExpression(expr, "b")

    expr = ScalarExpression("sin(a)", ["a", "b"])
    assert expr == ScalarExpression(expr)
    assert expr != ScalarExpression(expr, "a")
    with pytest.raises(RuntimeError):
        ScalarExpression(expr, "b")


def test_expression_user_funcs():
    """test the usage of user_funcs"""
    expr = ScalarExpression("func()", user_funcs={"func": lambda: 1})
    assert expr() == 1
    assert expr.get_compiled()() == 1
    assert expr.value == 1

    expr = ScalarExpression("f(pi)", user_funcs={"f": np.sin})
    assert expr.constant
    assert expr() == pytest.approx(0)
    assert expr.get_compiled()() == pytest.approx(0)
    assert expr.value == pytest.approx(0)

    expr = TensorExpression("[0, f(pi)]", user_funcs={"f": np.sin})
    assert expr.constant
    np.testing.assert_allclose(expr(), np.array([0, 0]), atol=1e-14)
    np.testing.assert_allclose(expr.get_compiled()(), np.array([0, 0]), atol=1e-14)
    np.testing.assert_allclose(expr.value, np.array([0, 0]), atol=1e-14)


def test_complex_expression():
    """test expressions with complex numbers"""
    for s in ["sqrt(-1)", "I"]:
        expr = ScalarExpression(s)
        assert expr.complex
        assert expr.constant
        assert expr.value == pytest.approx(1j)

    expr = TensorExpression("[1, I]")
    assert expr.complex
    assert expr.constant
    assert expr.rank == 1
    assert expr.shape == (2,)
    np.testing.assert_allclose(expr.value, np.array([1, 1j]))

    expr = TensorExpression("[[1, -1], [I, 2]]")
    assert expr.complex
    assert expr.constant
    assert expr.rank == 2
    assert expr.shape == (2, 2)
    np.testing.assert_allclose(expr.value, np.array([[1, -1], [1j, 2]]))


@pytest.mark.parametrize(
    "expression, value",
    [("Heaviside(x)", 0.5), ("Heaviside(x, 0.75)", 0.75), ("heaviside(x, 0.75)", 0.75)],
)
def test_expression_heaviside(expression, value):
    """test special cases of expressions"""
    expr = ScalarExpression(expression)
    assert not expr.constant
    assert expr(-1) == 0
    assert expr(0) == value
    assert expr(1) == 1
    np.testing.assert_allclose(expr(np.array([-1, 0, 1])), np.array([0, value, 1]))

    f = expr.get_compiled()
    assert f(-1) == 0
    assert f(0) == value
    assert f(1) == 1
    np.testing.assert_allclose(f(np.array([-1, 0, 1])), np.array([0, value, 1]))


def test_expression_consts():
    """test the usage of consts"""
    expr = ScalarExpression("a", consts={"a": 1})
    assert expr.constant
    assert not expr.depends_on("a")
    assert expr() == 1
    assert expr.get_compiled()() == 1
    assert expr.value == 1

    expr = ScalarExpression("a + b", consts={"a": 1})
    assert not expr.constant
    assert not expr.depends_on("a") and expr.depends_on("b")
    assert expr(2) == 3
    assert expr.get_compiled()(2) == 3

    expr = ScalarExpression("a + b", consts={"a": np.array([1, 2])})
    assert not expr.constant
    np.testing.assert_allclose(expr(np.array([2, 3])), np.array([3, 5]))
    np.testing.assert_allclose(expr.get_compiled()(np.array([2, 3])), np.array([3, 5]))


def test_evaluate_func_scalar():
    """test the evaluate function with scalar fields"""
    grid = UnitGrid([2, 4])
    field = ScalarField.from_expression(grid, "x")

    res1 = evaluate("laplace(a)", {"a": field})
    res2 = field.laplace("natural")
    np.testing.assert_almost_equal(res1.data, res2.data)

    res = evaluate("a - x", {"a": field})
    np.testing.assert_almost_equal(res.data, 0)

    field_c = ScalarField.from_expression(grid, "1 + 1j")
    res = evaluate("c", {"c": field_c})
    assert res.dtype == "complex"
    np.testing.assert_almost_equal(res.data, field_c.data)

    res = evaluate("abs(a)**2", {"a": field_c})
    assert res.dtype == "float"
    np.testing.assert_allclose(res.data, 2)

    # edge case
    res = evaluate("sin", {"a": field_c}, consts={"sin": 3.14})
    assert res.dtype == "float"
    np.testing.assert_allclose(res.data, 3.14)


def test_evaluate_func_vector():
    """test the evaluate function with vector fields"""
    grid = UnitGrid([3])
    field = ScalarField.from_expression(grid, "x")
    vec = VectorField.from_expression(grid, ["x"])

    res = evaluate("inner(v, v)", {"v": vec})
    assert isinstance(res, ScalarField)
    np.testing.assert_almost_equal(res.data, grid.axes_coords[0] ** 2)

    res = evaluate("outer(v, v)", {"v": vec})
    assert isinstance(res, Tensor2Field)
    np.testing.assert_almost_equal(res.data, [[grid.axes_coords[0] ** 2]])

    assert isinstance(evaluate("gradient(a)", {"a": field}), VectorField)


def test_evaluate_func_invalid():
    """test the evaluate function with invalid data"""
    field = ScalarField.from_expression(UnitGrid([3]), "x")

    with pytest.raises(ValueError):
        evaluate("x", {})

    with pytest.raises(ValueError):
        evaluate("x", {"x": field})

    field2 = ScalarField.from_expression(UnitGrid([4]), "x")
    with pytest.raises(ValueError):
        evaluate("a + b", {"a": field, "b": field2})

    with pytest.raises(BCDataError):
        evaluate("laplace(a)", {"a": field}, bc=RuntimeError)

    with pytest.raises(RuntimeError):
        evaluate("a + b", {"a": field})


def test_evaluate_func_bcs_warning(caplog):
    """test whether a warning is thrown correctly"""
    field = ScalarField.from_expression(UnitGrid([3]), "x")

    with caplog.at_level(logging.WARNING):
        evaluate("laplace(u)", {"u": field}, bc_ops={"u:gradient": "value"})
    assert "gradient" in caplog.text

    with caplog.at_level(logging.WARNING):
        evaluate("laplace(u)", {"u": field}, bc_ops={"v:laplace": "value"})
    assert "Unused" in caplog.text


def test_evaluate_func_collection():
    """test the evaluate function with a field collection"""
    grid = UnitGrid([3])
    field = ScalarField.from_expression(grid, "x")
    vec = VectorField.from_expression(grid, ["x"])
    col = FieldCollection({"v": vec, "a": field})

    res = evaluate("inner(v, v)", col)
    assert isinstance(res, ScalarField)
    np.testing.assert_almost_equal(res.data, grid.axes_coords[0] ** 2)

    assert isinstance(evaluate("gradient(a)", col), VectorField)
    assert isinstance(evaluate("a * v", col), VectorField)

    with pytest.raises(RuntimeError):
        col.labels = ["a", "a"]
        evaluate("1", col)



"""
Auxiliary mathematical functions

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

import math
from typing import Any, Dict, Optional, Tuple

import numba as nb
import numpy as np
from numba.experimental import jitclass

from .typing import ArrayLike


class SmoothData1D:
    """allows smoothing data in 1d using a Gaussian kernel of defined width

    The data is given a pairs of `x` and `y`, the assumption being that there is
    an underlying relation `y = f(x)`.
    """

    sigma_auto_scale: float = 10
    """float: scale for setting automatic values for sigma"""

    def __init__(self, x, y, sigma: Optional[float] = None):
        """initialize with data

        Args:
            x:
                List of x values
            y:
                List of y values
            sigma (float):
                The size of the smoothing window in units of `x`. If omitted, the
                average distance of x values multiplied by `sigma_auto_scale` is used.
        """
        self.x = np.ravel(x)
        self.y = np.ravel(y)
        if self.x.shape != self.y.shape:
            raise ValueError("`x` and `y` must have equal number of elements")

        # only take finite values
        idx = np.isfinite(self.y)
        if not np.all(idx):
            self.x = self.x[idx]
            self.y = self.y[idx]

        if sigma is None:
            self.sigma = float(self.sigma_auto_scale * self.x.ptp() / len(self.x))
        else:
            self.sigma = sigma

    @property
    def bounds(self) -> Tuple[float, float]:
        """return minimal and maximal `x` values"""
        return float(self.x.min()), float(self.x.max())

    def __contains__(self, x: float) -> bool:
        """checks whether the value `x` is contain in the range of x-values"""
        return self.x.min() <= x <= self.x.max()  # type: ignore

    def __call__(self, xs: ArrayLike) -> np.ndarray:
        """return smoothed y values for the positions given in `xs`

        Args:
            xs (list of :class:`~numpy.ndarray`): the x-values

        Returns:
            :class:`~numpy.ndarray`: The associated function values
        """
        xs = np.asanyarray(xs)
        shape = xs.shape
        xs = np.ravel(xs)
        scale = 0.5 * self.sigma**-2

        # determine the weights of all input points
        with np.errstate(under="ignore"):
            weight = np.exp(-scale * (self.x[:, None] - xs[None, :]) ** 2)
            weight_sum = weight.sum(axis=0)
            i = weight_sum > 0
            weight[..., i] /= weight_sum[i]

        result = self.y @ weight
        return result.reshape(shape)  # type: ignore

    def derivative(self, xs: ArrayLike) -> np.ndarray:
        """return the derivative of the smoothed values for the positions `xs`

        Note that this value

        Args:
            xs (list of :class:`~numpy.ndarray`): the x-values

        Returns:
            :class:`~numpy.ndarray`: The associated values of the derivative
        """
        xs = np.asanyarray(xs)
        shape = xs.shape
        xs = np.ravel(xs)
        scale = 0.5 * self.sigma**-2

        # determine the weights of all input points
        with np.errstate(under="ignore"):
            ws = np.exp(-scale * (self.x[:, None] - xs[None, :]) ** 2)
            # first axis varies internal points where we know values
            # second axis gives positions at which we want to evaluate values
            ws_sum = ws.sum(axis=0)
            i = ws_sum > 0
            ws[..., i] /= ws_sum[i]

        diff = 2 * scale * (self.x[:, None] - xs[None, :])
        result = self.y @ (diff * ws) - (self.y @ ws) * (diff * ws).sum(axis=0)
        return result.reshape(shape)  # type: ignore


@jitclass(
    [
        ("min", nb.double),
        ("max", nb.double),
        ("mean", nb.double),
        ("_mean2", nb.double),
        ("count", nb.uint),
    ]
)
class OnlineStatistics:
    """class for using an online algorithm for calculating statistics"""

    mean: float
    """float: recorded mean"""
    count: int
    """int: recorded number of items"""

    def __init__(self) -> None:
        self.min = np.inf
        self.max = -np.inf
        self.mean = 0
        self._mean2: float = 0
        self.count = 0

    @property
    def var(self) -> float:
        """float: recorded variance"""
        DDOF = 0
        if self.count <= DDOF:
            return math.nan
        else:
            return self._mean2 / (self.count - DDOF)

    @property
    def std(self) -> float:
        """float: recorded standard deviation"""
        return np.sqrt(self.var)  # type: ignore

    def add(self, value: float) -> None:
        """add a value to the accumulator

        Args:
            value (float): The value to add
        """
        self.min = min(self.min, value)
        self.max = max(self.max, value)
        delta = value - self.mean
        self.count += 1
        self.mean += delta / self.count
        self._mean2 += delta * (value - self.mean)

    def to_dict(self) -> Dict[str, Any]:
        """return the information as a dictionary"""
        return {
            "min": self.min,
            "max": self.max,
            "mean": self.mean,
            "std": self.std,
            "count": self.count,
        }



"""
Defines an implicit solver

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

import numba as nb
import numpy as np

from ..fields.base import FieldBase
from ..pdes.base import PDEBase
from ..tools.typing import BackendType
from .base import SolverBase


class ConvergenceError(RuntimeError):
    pass


class ImplicitSolver(SolverBase):
    """class for solving partial differential equations implicitly"""

    name = "implicit"

    def __init__(
        self,
        pde: PDEBase,
        maxiter: int = 100,
        maxerror: float = 1e-4,
        backend: BackendType = "auto",
    ):
        """
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            maxiter (int):
                The maximal number of iterations per step
            maxerror (float):
                The maximal error that is permitted in each step
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.
        """
        super().__init__(pde, backend=backend)
        self.maxiter = maxiter
        self.maxerror = maxerror

    def _make_single_step_fixed_dt(
        self, state: FieldBase, dt: float
    ) -> Callable[[np.ndarray, float], None]:
        """return a function doing a single step with an implicit Euler scheme

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the implicit step
        """
        if self.pde.is_sde:
            raise RuntimeError("Cannot use implicit stepper with stochastic equation")

        self.info["function_evaluations"] = 0
        self.info["scheme"] = "implicit-euler"
        self.info["stochastic"] = False
        self.info["dt_adaptive"] = False

        rhs = self._make_pde_rhs(state, backend=self.backend)
        maxiter = int(self.maxiter)
        maxerror2 = self.maxerror**2

        # handle deterministic version of the pde
        def implicit_step(state_data: np.ndarray, t: float) -> None:
            """compiled inner loop for speed"""
            nfev = 0  # count function evaluations

            # estimate state at next time point
            evolution_last = dt * rhs(state_data, t)

            for n in range(maxiter):
                # fixed point iteration for improving state after dt
                state_guess = state_data + evolution_last
                evolution_this = dt * rhs(state_guess, t + dt)

                # calculate mean squared error
                err = 0.0
                for j in range(state_data.size):
                    diff = (
                        state_guess.flat[j]
                        - state_data.flat[j]
                        - evolution_this.flat[j]
                    )
                    err += (diff.conjugate() * diff).real
                err /= state_data.size

                if err < maxerror2:
                    # fix point iteration converged
                    break

                evolution_last = evolution_this
            else:
                with nb.objmode:
                    self._logger.warning(
                        "Implicit Euler step did not converge after %d iterations "
                        "at t=%g (error=%g)",
                        maxiter,
                        t,
                        err,
                    )
                raise ConvergenceError("Implicit Euler step did not converge.")
            nfev += n + 1
            state_data += evolution_this

        return implicit_step



"""
Defines an explicit solver supporting various methods
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable, Literal, Optional, Tuple

import numba as nb
import numpy as np

from ..fields.base import FieldBase
from ..pdes.base import PDEBase
from ..tools.math import OnlineStatistics
from ..tools.numba import jit
from ..tools.typing import BackendType
from .base import AdaptiveSolverBase


class ExplicitSolver(AdaptiveSolverBase):
    """class for solving partial differential equations explicitly"""

    name = "explicit"

    def __init__(
        self,
        pde: PDEBase,
        scheme: Literal["euler", "runge-kutta", "rk", "rk45"] = "euler",
        *,
        backend: BackendType = "auto",
        adaptive: bool = False,
        tolerance: float = 1e-4,
    ):
        """
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            scheme (str):
                Defines the explicit scheme to use. Supported values are 'euler' and
                'runge-kutta' (or 'rk' for short).
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.
            adaptive (bool):
                When enabled, the time step is adjusted during the simulation using the
                error tolerance set with `tolerance`.
            tolerance (float):
                The error tolerance used in adaptive time stepping. This is used in
                adaptive time stepping to choose a time step which is small enough so
                the truncation error of a single step is below `tolerance`.
        """
        super().__init__(pde, backend=backend, adaptive=adaptive, tolerance=tolerance)
        self.scheme = scheme

    def _make_single_step_fixed_euler(
        self, state: FieldBase, dt: float
    ) -> Callable[[np.ndarray, float], None]:
        """make a simple Euler stepper with fixed time step

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the explicit stepping.

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, steps: int)`
        """
        if self.pde.is_sde:
            # handle stochastic version of the pde
            rhs_sde = self._make_sde_rhs(state, backend=self.backend)

            def stepper(state_data: np.ndarray, t: float) -> None:
                """perform a single Euler-Maruyama step"""
                evolution_rate, noise_realization = rhs_sde(state_data, t)
                state_data += dt * evolution_rate
                if noise_realization is not None:
                    state_data += np.sqrt(dt) * noise_realization

            self._logger.info("Init explicit Euler-Maruyama stepper with dt=%g", dt)

        else:
            # handle deterministic version of the pde
            rhs_pde = self._make_pde_rhs(state, backend=self.backend)

            def stepper(state_data: np.ndarray, t: float) -> None:
                """perform a single Euler step"""
                state_data += dt * rhs_pde(state_data, t)

            self._logger.info("Init explicit Euler stepper with dt=%g", dt)

        return stepper

    def _make_single_step_rk45(
        self, state: FieldBase, dt: float
    ) -> Callable[[np.ndarray, float], None]:
        """make function doing a single explicit Runge-Kutta step of order 5(4)

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the explicit stepping.

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, steps: int)`
        """
        if self.pde.is_sde:
            raise RuntimeError("Runge-Kutta stepper does not support stochasticity")

        # obtain functions determining how the PDE is evolved
        rhs = self._make_pde_rhs(state, backend=self.backend)

        def stepper(state_data: np.ndarray, t: float) -> None:
            """compiled inner loop for speed"""
            # calculate the intermediate values in Runge-Kutta
            k1 = dt * rhs(state_data, t)
            k2 = dt * rhs(state_data + 0.5 * k1, t + 0.5 * dt)
            k3 = dt * rhs(state_data + 0.5 * k2, t + 0.5 * dt)
            k4 = dt * rhs(state_data + k3, t + dt)

            state_data += (k1 + 2 * k2 + 2 * k3 + k4) / 6

        self._logger.info("Init explicit Runge-Kutta-45 stepper with dt=%g", dt)
        return stepper

    def _make_single_step_fixed_dt(
        self, state: FieldBase, dt: float
    ) -> Callable[[np.ndarray, float], None]:
        """return a function doing a single step with a fixed time step

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the explicit stepping.
        """
        self.info["scheme"] = self.scheme
        if self.scheme == "euler":
            return self._make_single_step_fixed_euler(state, dt)
        elif self.scheme in {"runge-kutta", "rk", "rk45"}:
            return self._make_single_step_rk45(state, dt)
        else:
            raise ValueError(f"Explicit scheme `{self.scheme}` is not supported")

    def _make_adaptive_euler_stepper(
        self, state: FieldBase
    ) -> Callable[
        [np.ndarray, float, float, float, Optional[OnlineStatistics]],
        Tuple[float, float, int, float],
    ]:
        """make an adaptive Euler stepper

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, t_end: float)`
        """
        # General comment: We implement the full adaptive scheme here instead of just
        # defining `_make_single_step_error_estimate` to do some optimizations. In
        # particular, we reuse the calculated right hand side in cases where the step
        # was not succesful.
        if self.pde.is_sde:
            raise RuntimeError("Cannot use adaptive stepper with stochastic equation")

        # obtain functions determining how the PDE is evolved
        rhs_pde = self._make_pde_rhs(state, backend=self.backend)
        modify_after_step = jit(self.pde.make_modify_after_step(state))

        # obtain auxiliary functions
        sync_errors = self._make_error_synchronizer()
        adjust_dt = self._make_dt_adjuster()
        tolerance = self.tolerance
        dt_min = self.dt_min

        def adaptive_stepper(
            state_data: np.ndarray,
            t_start: float,
            t_end: float,
            dt_init: float,
            dt_stats: Optional[OnlineStatistics] = None,
        ) -> Tuple[float, float, int, float]:
            """adaptive stepper that advances the state in time"""
            modifications = 0.0
            dt_opt = dt_init
            rate = rhs_pde(state_data, t_start)  # calculate initial rate

            steps = 0
            t = t_start
            while True:
                # use a smaller (but not too small) time step if close to t_end
                dt_step = max(min(dt_opt, t_end - t), dt_min)

                # do single step with dt
                step_large = state_data + dt_step * rate

                # do double step with half the time step
                step_small = state_data + 0.5 * dt_step * rate

                try:
                    # calculate rate at the midpoint of the double step
                    rate_midpoint = rhs_pde(step_small, t + 0.5 * dt_step)
                except Exception:
                    # an exception likely signals that rate could not be calculated
                    error_rel = np.nan
                else:
                    # advance to end of double step
                    step_small += 0.5 * dt_step * rate_midpoint

                    # calculate maximal error
                    error = np.abs(step_large - step_small).max()
                    error_rel = error / tolerance  # normalize error to given tolerance

                # synchronize the error between all processes (if necessary)
                error_rel = sync_errors(error_rel)

                if error_rel <= 1:  # error is sufficiently small
                    try:
                        # calculating the rate at putative new step
                        rate = rhs_pde(step_small, t)
                    except Exception:
                        # calculating the rate failed => retry with smaller dt
                        error_rel = np.nan
                    else:
                        # everything worked => do the step
                        steps += 1
                        t += dt_step
                        state_data[...] = step_small
                        modifications += modify_after_step(state_data)
                        if dt_stats is not None:
                            dt_stats.add(dt_step)

                if t < t_end:
                    # adjust the time step and continue
                    dt_opt = adjust_dt(dt_step, error_rel)
                else:
                    break  # return to the controller

            return t, dt_opt, steps, modifications

        if self._compiled:
            # compile inner stepper
            sig_adaptive = (
                nb.typeof(state.data),
                nb.double,
                nb.double,
                nb.double,
                nb.typeof(self.info["dt_statistics"]),
            )
            adaptive_stepper = jit(sig_adaptive)(adaptive_stepper)

        self._logger.info(f"Init adaptive Euler stepper")
        return adaptive_stepper

    def _make_single_step_error_estimate_rkf(
        self, state: FieldBase
    ) -> Callable[[np.ndarray, float, float], Tuple[np.ndarray, float]]:
        """make an adaptive stepper using the explicit Runge-Kutta-Fehlberg method

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, t_end: float)`
        """
        if self.pde.is_sde:
            raise RuntimeError("Cannot use adaptive stepper with stochastic equation")

        # obtain functions determining how the PDE is evolved
        rhs = self._make_pde_rhs(state, backend=self.backend)

        # use Runge-Kutta-Fehlberg method
        # define coefficients for RK4(5), formula 2 Table III in Fehlberg
        a2 = 1 / 4
        a3 = 3 / 8
        a4 = 12 / 13
        a5 = 1.0
        a6 = 1 / 2

        b21 = 1 / 4
        b31 = 3 / 32
        b32 = 9 / 32
        b41 = 1932 / 2197
        b42 = -7200 / 2197
        b43 = 7296 / 2197
        b51 = 439 / 216
        b52 = -8.0
        b53 = 3680 / 513
        b54 = -845 / 4104
        b61 = -8 / 27
        b62 = 2.0
        b63 = -3544 / 2565
        b64 = 1859 / 4104
        b65 = -11 / 40

        r1 = 1 / 360
        # r2 = 0
        r3 = -128 / 4275
        r4 = -2197 / 75240
        r5 = 1 / 50
        r6 = 2 / 55

        c1 = 25 / 216
        # c2 = 0
        c3 = 1408 / 2565
        c4 = 2197 / 4104
        c5 = -1 / 5

        def stepper(
            state_data: np.ndarray, t: float, dt: float
        ) -> Tuple[np.ndarray, float]:
            """basic stepper to estimate error"""
            # do the six intermediate steps
            k1 = dt * rhs(state_data, t)
            k2 = dt * rhs(state_data + b21 * k1, t + a2 * dt)
            k3 = dt * rhs(state_data + b31 * k1 + b32 * k2, t + a3 * dt)
            k4 = dt * rhs(state_data + b41 * k1 + b42 * k2 + b43 * k3, t + a4 * dt)
            k5 = dt * rhs(
                state_data + b51 * k1 + b52 * k2 + b53 * k3 + b54 * k4,
                t + a5 * dt,
            )
            k6 = dt * rhs(
                state_data + b61 * k1 + b62 * k2 + b63 * k3 + b64 * k4 + b65 * k5,
                t + a6 * dt,
            )

            # estimate the maximal error
            error_local = r1 * k1 + r3 * k3 + r4 * k4 + r5 * k5 + r6 * k6
            error = np.abs(error_local).max()

            state_new = state_data + c1 * k1 + c3 * k3 + c4 * k4 + c5 * k5
            return state_new, error

        self._logger.info("Init adaptive Runge-Kutta-Fehlberg stepper")
        return stepper

    def _make_single_step_error_estimate(
        self, state: FieldBase
    ) -> Callable[[np.ndarray, float, float], Tuple[np.ndarray, float]]:
        """make a stepper that also estimates the error

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
        """
        self.info["scheme"] = self.scheme
        if self.scheme in {"runge-kutta", "rk", "rk45"}:
            return self._make_single_step_error_estimate_rkf(state)
        else:
            # Note that the Euler scheme is implemented separately to use detailed
            # optimizations; see method `_make_adaptive_euler_stepper`
            raise ValueError(f"Adaptive scheme `{self.scheme}` is not supported")

    def _make_adaptive_stepper(
        self, state: FieldBase
    ) -> Callable[
        [np.ndarray, float, float, float, Optional[OnlineStatistics]],
        Tuple[float, float, int, float],
    ]:
        """return a stepper function using an explicit scheme with fixed time steps

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
        """
        if self.pde.is_sde:
            raise RuntimeError("Cannot use adaptive stepper with stochastic equation")

        self.info["scheme"] = self.scheme
        if self.scheme == "euler":
            return self._make_adaptive_euler_stepper(state)
        elif self.scheme in {"runge-kutta", "rk", "rk45"}:
            return super()._make_adaptive_stepper(state)
        else:
            raise ValueError(f"Adaptive scheme `{self.scheme}` is not supported")



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import ScalarField, UnitGrid, pdes


@pytest.mark.parametrize("dim", [1, 2])
@pytest.mark.parametrize(
    "pde_class",
    [
        pdes.AllenCahnPDE,
        pdes.CahnHilliardPDE,
        pdes.DiffusionPDE,
        pdes.KPZInterfacePDE,
        pdes.KuramotoSivashinskyPDE,
        pdes.SwiftHohenbergPDE,
    ],
)
def test_pde_consistency(pde_class, dim, rng):
    """test some methods of generic PDE models"""
    eq = pde_class()
    assert isinstance(str(eq), str)
    assert isinstance(repr(eq), str)

    # compare numba to numpy implementation
    grid = UnitGrid([4] * dim)
    state = ScalarField.random_uniform(grid, rng=rng)
    field = eq.evolution_rate(state)
    assert field.grid == grid
    rhs = eq._make_pde_rhs_numba(state)
    res = rhs(state.data, 0)
    np.testing.assert_allclose(field.data, res)

    # compare to generic implementation
    assert isinstance(eq.expression, str)
    eq2 = pdes.PDE({"c": eq.expression})
    np.testing.assert_allclose(field.data, eq2.evolution_rate(state).data)


def test_pde_consistency_test(rng):
    """test whether the consistency of a pde implementation is checked"""

    class TestPDE(pdes.PDEBase):
        def evolution_rate(self, field, t=0):
            return 2 * field

        def _make_pde_rhs_numba(self, state):
            def impl(state_data, t):
                return 3 * state_data

            return impl

    eq = TestPDE()
    state = ScalarField.random_uniform(UnitGrid([4]), rng=rng)
    with pytest.raises(AssertionError):
        eq.solve(state, t_range=5, tracker=None)



"""
Classes for tracking simulation results in controlled interrupts

Trackers are classes that periodically receive the state of the simulation to analyze,
store, or output it. The trackers defined in this module are:

.. autosummary::
   :nosignatures:

   ~trackers.CallbackTracker
   ~trackers.ProgressTracker
   ~trackers.PrintTracker
   ~trackers.PlotTracker
   ~trackers.LivePlotTracker
   ~trackers.DataTracker
   ~trackers.SteadyStateTracker
   ~trackers.RuntimeTracker
   ~trackers.ConsistencyTracker
   ~interactive.InteractivePlotTracker
   
Some trackers can also be referenced by name for convenience when using them in
simulations. The lit of supported names is returned by
:func:`~pde.trackers.base.get_named_trackers`.
   
Multiple trackers can be collected in a :class:`~base.TrackerCollection`, which provides
methods for handling them efficiently. Moreover, custom trackers can be implemented by
deriving from :class:`~.trackers.base.TrackerBase`. Note that trackers generally receive
a view into the current state, implying that they can adjust the state by modifying it
in-place. Moreover, trackers can interrupt the simulation by raising the special
exception :class:`StopIteration`.


For each tracker, the time intervals at which it is called can be decided using one
of the following classes, which determine when the simulation will be interrupted:

.. autosummary::
   :nosignatures:
   
   ~interrupts.FixedInterrupts
   ~interrupts.ConstantInterrupts
   ~interrupts.LogarithmicInterrupts
   ~interrupts.RealtimeInterrupts
   
In particular, interrupts can be specified conveniently using
:func:`~interrupts.interval_to_interrupts`.
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""


from .base import get_named_trackers
from .interactive import InteractivePlotTracker
from .interrupts import (
    ConstantInterrupts,
    FixedInterrupts,
    LogarithmicInterrupts,
    RealtimeInterrupts,
    interval_to_interrupts,
)
from .trackers import *



"""
Cartesian grids of arbitrary dimension.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from __future__ import annotations

import itertools
from typing import (  # @UnusedImport
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from ..tools.cuboid import Cuboid
from ..tools.plotting import plot_on_axes
from .base import CoordsType, DimensionError, GridBase, _check_shape

if TYPE_CHECKING:
    from .boundaries.axes import Boundaries, BoundariesData  # @UnusedImport


class CartesianGrid(GridBase):
    r""" d-dimensional Cartesian grid with uniform discretization for each axis

    The grids can be thought of as a collection of n-dimensional boxes, called cells, of
    equal length in each dimension. The bounds then defined the total volume covered by
    these cells, while the cell coordinates give the location of the box centers. We
    index the boxes starting from 0 along each dimension. Consequently, the cell
    :math:`i-\frac12` corresponds to the left edge of the  covered interval and the
    index :math:`i+\frac12` corresponds to the right edge, when the dimension is covered
    by d boxes.

    In particular, the discretization along dimension :math:`k` is defined as

    .. math::
            x^{(k)}_i &= x^{(k)}_\mathrm{min} + \left(i + \frac12\right)
                \Delta x^{(k)}
            \quad \text{for} \quad i = 0, \ldots, N^{(k)} - 1
        \\
            \Delta x^{(k)} &= \frac{x^{(k)}_\mathrm{max} -
                                    x^{(k)}_\mathrm{min}}{N^{(k)}}

    where :math:`N^{(k)}` is the number of cells along this dimension. Consequently, 
    cells have dimension :math:`\Delta x^{(k)}` and cover the interval
    :math:`[x^{(k)}_\mathrm{min}, x^{(k)}_\mathrm{max}]`.
    """

    cuboid: Cuboid

    boundary_names = {  # name all the boundaries
        "left": (0, False),
        "right": (0, True),
        "bottom": (1, False),
        "top": (1, True),
        "back": (2, False),
        "front": (2, True),
    }

    def __init__(
        self,
        bounds: Sequence[Tuple[float, float]],
        shape: Union[int, Sequence[int]],
        periodic: Union[Sequence[bool], bool] = False,
    ):
        """
        Args:
            bounds (list of tuple):
                Give the coordinate range for each axis. This should be a tuple of two
                number (lower and upper bound) for each axis. The length of `bounds`
                thus determines the grid dimension.
            shape (list):
                The number of support points for each axis. The length of `shape` needs
                to match the grid dimension.
            periodic (bool or list):
                Specifies which axes possess periodic boundary conditions. This is
                either a list of booleans defining periodicity for each individual axis
                or a single boolean value specifying the same periodicity for all axes.
        """
        bounds_arr = np.array(bounds, ndmin=1, dtype=np.double)
        if bounds_arr.shape == (2,):
            raise ValueError(
                "`bounds with shape (2,) are ambiguous. Either use shape (1, 2) to set "
                "up a 1d system with two bounds or shape (2, 1) for a 2d system with "
                "only the upper bounds specified"
            )

        if bounds_arr.ndim == 1 or bounds_arr.shape[1] == 1:
            # only set the upper bounds
            bounds_arr = np.atleast_1d(np.squeeze(bounds_arr))
            self.cuboid = Cuboid(np.zeros_like(bounds_arr), bounds_arr, mutable=False)

        elif bounds_arr.ndim == 2 and bounds_arr.shape[1] == 2:
            # upper and lower bounds of the grid are given
            self.cuboid = Cuboid.from_bounds(bounds_arr, mutable=False)

        else:
            raise ValueError(
                f"Do not know how to interpret shape {bounds_arr.shape} for bounds"
            )

        # handle the shape array
        shape = _check_shape(shape)
        if len(shape) == 1 and self.cuboid.dim > 1:
            shape = (int(shape[0]),) * self.cuboid.dim
        if self.cuboid.dim != len(shape):
            raise DimensionError("Dimension of `bounds` and `shape` are not compatible")

        # initialize the base class
        super().__init__()
        self._shape = _check_shape(shape)
        self.dim = len(self.shape)
        self.num_axes = self.dim

        if isinstance(periodic, (bool, np.bool_)):
            self._periodic = [bool(periodic)] * self.dim
        elif len(periodic) != self.dim:
            raise DimensionError(
                "Number of axes with specified periodicity does not match grid "
                f"dimension ({len(periodic)} != {self.dim})"
            )
        else:
            self._periodic = list(periodic)

        if self.dim <= 3:
            self.axes = list("xyz"[: self.dim])
        else:
            self.axes = [chr(97 + i) for i in range(self.dim)]

        # determine the coordinates
        p1, p2 = self.cuboid.corners
        axes_coords, discretization = [], []
        for d in range(self.dim):
            num = self.shape[d]
            c, dc = np.linspace(p1[d], p2[d], num, endpoint=False, retstep=True)
            if self.shape[d] == 1:
                # correct for singular dimension
                dc = p2[d] - p1[d]
            c += dc / 2
            axes_coords.append(c)
            discretization.append(dc)
        self._discretization = np.array(discretization)
        self._axes_coords = tuple(axes_coords)
        self._axes_bounds = tuple(self.cuboid.bounds)

    @property
    def state(self) -> Dict[str, Any]:
        """dict: the state of the grid"""
        return {
            "bounds": self.axes_bounds,
            "shape": self.shape,
            "periodic": self.periodic,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> CartesianGrid:  # type: ignore
        """create a field from a stored `state`.

        Args:
            state (dict):
                The state from which the grid is reconstructed.

        Returns:
            :class:`CartesianGrid`: the grid re-created from the state data
        """
        state_copy = state.copy()
        obj = cls(
            bounds=state_copy.pop("bounds"),
            shape=state_copy.pop("shape"),
            periodic=state_copy.pop("periodic"),
        )
        if state_copy:
            raise ValueError(f"State items {state_copy.keys()} were not used")
        return obj

    @classmethod
    def from_bounds(
        cls,
        bounds: Sequence[Tuple[float, float]],
        shape: Sequence[int],
        periodic: Sequence[bool],
    ) -> CartesianGrid:
        """
        Args:
            bounds (tuple):
                Give the coordinate range for each axis. This should be a tuple of two
                number (lower and upper bound) for each axis. The length of `bounds`
                thus determines the grid dimension.
            shape (tuple):
                The number of support points for each axis. The length of `shape` needs
                to match the grid dimension.
            periodic (bool or list):
                Specifies which axes possess periodic boundary conditions. This is
                either a list of booleans defining periodicity for each individual axis
                or a single boolean value specifying the same periodicity for all axes.

        Returns:
            :class:`CartesianGrid`: representing the region chosen by bounds
        """
        return CartesianGrid(bounds, shape, periodic)

    @property
    def volume(self) -> float:
        """float: total volume of the grid"""
        return float(self.cuboid.volume)

    @property
    def cell_volume_data(self):
        """size associated with each cell"""
        return tuple(self.discretization)

    def iter_mirror_points(
        self, point: np.ndarray, with_self: bool = False, only_periodic: bool = True
    ) -> Generator:
        """generates all mirror points corresponding to `point`

        Args:
            point (:class:`~numpy.ndarray`):
                The point within the grid
            with_self (bool):
                Whether to include the point itself
            only_periodic (bool):
                Whether to only mirror along periodic axes

        Returns:
            A generator yielding the coordinates that correspond to mirrors
        """
        point = np.asanyarray(point, dtype=np.double)

        # find all offsets of the individual axes
        offsets = []
        for i in range(self.dim):
            if only_periodic and not self.periodic[i]:
                offsets.append([0])
            else:
                s = self.cuboid.size[i]
                offsets.append([-s, 0, s])

        # produce the respective mirrored points
        for offset in itertools.product(*offsets):
            if with_self or np.linalg.norm(offset) != 0:
                yield point + offset

    def get_random_point(
        self,
        *,
        boundary_distance: float = 0,
        coords: CoordsType = "cartesian",
        rng: Optional[np.random.Generator] = None,
    ) -> np.ndarray:
        """return a random point within the grid

        Args:
            boundary_distance (float):
                The minimal distance this point needs to have from all boundaries.
            coords (str):
                Determines the coordinate system in which the point is specified. Valid
                values are `cartesian`, `cell`, and `grid`;
                see :meth:`~pde.grids.base.GridBase.transform`.
            rng (:class:`~numpy.random.Generator`):
                Random number generator (default: :func:`~numpy.random.default_rng()`)

        Returns:
            :class:`~numpy.ndarray`: The coordinates of the point
        """
        if rng is None:
            rng = np.random.default_rng()

        # handle the boundary distance
        cuboid = self.cuboid
        if boundary_distance != 0:
            if any(cuboid.size <= 2 * boundary_distance):
                raise RuntimeError("Random points would be too close to boundary")
            cuboid = cuboid.buffer(-boundary_distance)

        # create random point
        point = cuboid.pos + rng.random(self.dim) * cuboid.size

        if coords == "cartesian" or coords == "grid":
            return point  # type: ignore
        elif coords == "cell":
            return self.transform(point, "grid", "cell")
        else:
            raise ValueError(f"Unknown coordinate system `{coords}`")

    def get_line_data(self, data: np.ndarray, extract: str = "auto") -> Dict[str, Any]:
        """return a line cut through the given data

        Args:
            data (:class:`~numpy.ndarray`):
                The values at the grid points
            extract (str):
                Determines which cut is done through the grid. Possible choices
                are (default is `cut_0`):

                * `cut_#`: return values along the axis specified by # and use
                  the mid point along all other axes.
                * `project_#`: average values for all axes, except axis #.

                Here, # can either be a zero-based index (from 0 to dim-1) or
                a letter denoting the axis.

        Returns:
            dict: A dictionary with information about the line cut, which is convenient
            for plotting.
        """
        if data.shape[-self.dim :] != self.shape:
            raise ValueError(
                f"Shape {data.shape} of the data array is not compatible with grid "
                f"shape {self.shape}"
            )

        def _get_axis(axis):
            """determine the axis from a given specifier"""
            try:
                axis = int(axis)
            except ValueError:
                try:
                    axis = self.axes.index(axis)
                except ValueError:
                    raise ValueError(f"Axis `{axis}` not defined")
            return axis

        if extract == "auto":
            extract = "cut_0"  # use a cut along first axis

        if extract.startswith("cut_"):
            # consider a cut along a given axis
            axis = _get_axis(extract[4:])
            data_y = data
            rank = data.ndim - self.dim  # rank of data
            for ax in reversed(range(self.dim)):
                if ax != axis:
                    mid_point = self.shape[ax] // 2
                    data_y = np.take(data_y, mid_point, axis=ax + rank)
            label_y = f"Cut along {self.axes[axis]}"

        elif extract.startswith("project_"):
            # consider a projection along a given axis
            axis = _get_axis(extract[8:])
            avg_axes = [ax - self.dim for ax in range(self.dim) if ax != axis]
            data_y = data.mean(axis=tuple(avg_axes))
            label_y = f"Projection onto {self.axes[axis]}"

        else:
            raise ValueError(f"Unknown extraction method `{extract}`")

        if self.dim == 1:
            label_y = ""

        # return the data with the respective labels
        return {
            "data_x": self.axes_coords[axis],
            "data_y": data_y,
            "extent_x": self.axes_bounds[axis],
            "label_x": self.axes[axis],
            "label_y": label_y,
        }

    def get_image_data(self, data: np.ndarray) -> Dict[str, Any]:
        """return a 2d-image of the data

        Args:
            data (:class:`~numpy.ndarray`):
                The values at the grid points

        Returns:
            dict: A dictionary with information about the image, which is  convenient
            for plotting.
        """
        if data.shape[-self.dim :] != self.shape:
            raise ValueError(
                f"Shape {data.shape} of the data array is not compatible with grid "
                f"shape {self.shape}"
            )

        if self.dim == 2:
            image_data = data
        elif self.dim == 3:
            image_data = data[:, :, self.shape[-1] // 2]
        else:
            raise NotImplementedError(
                "Creating images is only implemented for 2d and 3d grids"
            )

        extent: List[float] = []
        for c in self.axes_bounds[:2]:
            extent.extend(c)

        return {
            "data": image_data,
            "x": self.axes_coords[0],
            "y": self.axes_coords[1],
            "extent": extent,
            "label_x": self.axes[0],
            "label_y": self.axes[1],
        }

    def point_to_cartesian(
        self, points: np.ndarray, *, full: bool = False
    ) -> np.ndarray:
        """convert coordinates of a point to Cartesian coordinates

        Args:
            points (:class:`~numpy.ndarray`):
                Points given in grid coordinates
            full (bool):
                Compatibility option not used in this method

        Returns:
            :class:`~numpy.ndarray`: The Cartesian coordinates of the point
        """
        assert points.shape[-1] == self.dim, f"Point must have {self.dim} coordinates"
        return points

    def point_from_cartesian(self, coords: np.ndarray) -> np.ndarray:
        """convert points given in Cartesian coordinates to this grid

        Args:
            coords (:class:`~numpy.ndarray`):
                Points in Cartesian coordinates.

        Returns:
            :class:`~numpy.ndarray`: Points given in the coordinates of the grid
        """
        assert coords.shape[-1] == self.dim, f"Point must have {self.dim} coordinates"
        return coords

    def polar_coordinates_real(
        self, origin: np.ndarray, *, ret_angle: bool = False
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]]:
        """return polar coordinates associated with the grid

        Args:
            origin (:class:`~numpy.ndarray`):
                Coordinates of the origin at which the polar coordinate system is
                anchored.
            ret_angle (bool):
                Determines whether angles are returned alongside the distance. If `False`
                only the distance to the origin is returned for each support point of the
                grid. If `True`, the distance and angles are returned. For a 1d system
                system, the angle is defined as the sign of the difference between the
                point and the origin, so that angles can either be 1 or -1. For 2d
                systems and 3d systems, polar coordinates and spherical coordinates are
                used, respectively.

        Returns:
            :class:`~numpy.ndarray` or tuple of :class:`~numpy.ndarray`:
                Coordinates values in polar coordinates
        """
        origin = np.array(origin, dtype=np.double, ndmin=1)
        if len(origin) != self.dim:
            raise DimensionError("Dimensions are not compatible")

        # calculate the difference vector between all cells and the origin
        diff = self.difference_vector_real(origin, self.cell_coords)
        dist: np.ndarray = np.linalg.norm(diff, axis=-1)  # get distance

        # determine distance and optionally angles for these vectors
        if ret_angle:
            if self.dim == 1:
                return dist, np.sign(diff)[..., 0]  # type: ignore

            elif self.dim == 2:
                return dist, np.arctan2(diff[:, :, 0], diff[:, :, 1])  # type: ignore

            elif self.dim == 3:
                theta = np.arccos(diff[..., 2] / dist)
                phi = np.arctan2(diff[..., 0], diff[..., 1])
                return dist, theta, phi

            else:
                raise NotImplementedError(
                    f"Cannot calculate angles for dimension {self.dim}"
                )
        else:
            return dist

    def from_polar_coordinates(
        self,
        distance: np.ndarray,
        angle: np.ndarray,
        origin: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """convert polar coordinates to Cartesian coordinates

        This function is currently only implemented for 1d and 2d systems.

        Args:
            distance (:class:`~numpy.ndarray`):
                The radial distance
            angle (:class:`~numpy.ndarray`):
                The angle with respect to the origin
            origin (:class:`~numpy.ndarray`, optional):
                Sets the origin of the coordinate system. If omitted, the zero point is
                assumed as the origin.

        Returns:
            :class:`~numpy.ndarray`: The Cartesian coordinates corresponding to the given
            polar coordinates.
        """
        distance = np.asarray(distance)
        angle = np.asarray(angle)
        if origin is None:
            origin = np.zeros(self.dim)
        else:
            origin = np.atleast_1d(origin)

        if self.dim == 1:
            diff = distance * angle
            coords = origin + diff[..., None]

        elif self.dim == 2:
            unit_vector = np.moveaxis(np.array([np.sin(angle), np.cos(angle)]), 0, -1)
            diff = distance[..., None] * unit_vector
            coords = origin + diff

        else:
            raise NotImplementedError(
                f"Cannot calculate coordinates for dimension {self.dim}"
            )

        return self.normalize_point(coords, reflect=False)

    @plot_on_axes()
    def plot(self, ax, **kwargs):
        r"""visualize the grid

        Args:
            {PLOT_ARGS}
            \**kwargs: Extra arguments are passed on the to the matplotlib
                plotting routines, e.g., to set the color of the lines
        """
        if self.dim not in {1, 2}:
            raise NotImplementedError(
                f"Plotting is not implemented for grids of dimension {self.dim}"
            )

        kwargs.setdefault("color", "k")
        xb = self.axes_bounds[0]
        for x in np.linspace(*xb, self.shape[0] + 1):
            ax.axvline(x, **kwargs)
        ax.set_xlim(*xb)
        ax.set_xlabel(self.axes[0])

        if self.dim == 2:
            yb = self.axes_bounds[1]
            for y in np.linspace(*yb, self.shape[1] + 1):
                ax.axhline(y, **kwargs)
            ax.set_ylim(*yb)
            ax.set_ylabel(self.axes[1])

            ax.set_aspect(1)

    def slice(self, indices: Sequence[int]) -> CartesianGrid:
        """return a subgrid of only the specified axes

        Args:
            indices (list):
                Indices indicating the axes that are retained in the subgrid

        Returns:
            :class:`CartesianGrid`: The subgrid
        """
        subgrid = self.__class__(
            bounds=[self.axes_bounds[i] for i in indices],
            shape=tuple(self.shape[i] for i in indices),
            periodic=[self.periodic[i] for i in indices],
        )
        subgrid.axes = [self.axes[i] for i in indices]
        return subgrid


class UnitGrid(CartesianGrid):
    r"""d-dimensional Cartesian grid with unit discretization in all directions

    The grids can be thought of as a collection of d-dimensional cells of unit length.
    The `shape` parameter determines how many boxes there are in each direction. The
    cells are enumerated starting with 0, so the last cell has index :math:`n-1` if
    there are :math:`n` cells along a dimension. A given cell :math:`i` extends from
    coordinates :math:`i` to :math:`i + 1`, so the midpoint is at :math:`i + \frac12`,
    which is the cell coordinate. Taken together, the cells covers the interval
    :math:`[0, n]` along this dimension.
    """

    def __init__(
        self, shape: Sequence[int], periodic: Union[Sequence[bool], bool] = False
    ):
        """
        Args:
            shape (list):
                The number of support points for each axis. The dimension of the grid is
                given by `len(shape)`.
            periodic (bool or list):
                Specifies which axes possess periodic boundary conditions. This is
                either a list of booleans defining periodicity for each individual axis
                or a single boolean value specifying the same periodicity for all axes.
        """
        if isinstance(shape, int):
            shape = [shape]
        super().__init__([(0, s) for s in shape], shape, periodic)
        self.cuboid = Cuboid(np.zeros(self.dim), self.shape)
        self._discretization = np.ones(self.dim)

        # determine the cell center coordinates
        self._axes_coords = tuple(np.arange(n) + 0.5 for n in self.shape)
        self._axes_bounds = tuple(self.cuboid.bounds)

    @property
    def state(self) -> Dict[str, Any]:
        """dict: the state of the grid"""
        return {"shape": self.shape, "periodic": self.periodic}

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> UnitGrid:  # type: ignore
        """create a field from a stored `state`.

        Args:
            state (dict):
                The state from which the grid is reconstructed.
        """
        state_copy = state.copy()
        obj = cls(shape=state_copy.pop("shape"), periodic=state_copy.pop("periodic"))
        if state_copy:
            raise ValueError(f"State items {state_copy.keys()} were not used")
        return obj

    def to_cartesian(self) -> CartesianGrid:
        """convert unit grid to :class:`CartesianGrid`

        Returns:
            :class:`CartesianGrid`: The equivalent cartesian grid
        """
        return CartesianGrid(
            self.cuboid.bounds, shape=self.shape, periodic=self.periodic
        )

    def slice(self, indices: Sequence[int]) -> UnitGrid:
        """return a subgrid of only the specified axes

        Args:
            indices (list):
                Indices indicating the axes that are retained in the subgrid

        Returns:
            :class:`UnitGrid`: The subgrid
        """
        subgrid = self.__class__(
            shape=[self.shape[i] for i in indices],
            periodic=[self.periodic[i] for i in indices],
        )
        subgrid.axes = [self.axes[i] for i in indices]
        return subgrid



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from pde.tools import output


def test_progress_bars():
    """test progress bars"""
    pb_cls = output.get_progress_bar_class()
    tot = 0
    for i in pb_cls(range(4)):
        tot += i
    assert tot == 6


def test_in_jupyter_notebook():
    """test the function in_jupyter_notebook"""
    assert isinstance(output.in_jupyter_notebook(), bool)


def test_display_progress(capsys):
    """test whether this works"""
    for _ in output.display_progress(range(2)):
        pass
    out, err = capsys.readouterr()
    assert out == ""
    assert len(err) > 0



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import json
import os

import numpy as np
import pytest

from pde.tools import misc


def test_ensure_directory_exists(tmp_path):
    """tests the ensure_directory_exists function"""
    # create temporary name
    path = tmp_path / "test_ensure_directory_exists"
    assert not path.exists()
    # create the folder
    misc.ensure_directory_exists(path)
    assert path.is_dir()
    # check that a second call has the same result
    misc.ensure_directory_exists(path)
    assert path.is_dir()
    # remove the folder again
    os.rmdir(path)
    assert not path.exists()


def test_preserve_scalars():
    """test the preserve_scalars decorator"""

    class Test:
        @misc.preserve_scalars
        def meth(self, arr):
            return arr + 1

    t = Test()

    assert t.meth(1) == 2
    np.testing.assert_equal(t.meth(np.ones(2)), np.full(2, 2))


def test_hybridmethod():
    """test the hybridmethod decorator"""

    class Test:
        @misc.hybridmethod
        def method(cls):  # @NoSelf
            return "class"

        @method.instancemethod
        def method(self):
            return "instance"

    assert Test.method() == "class"
    assert Test().method() == "instance"


def test_estimate_computation_speed():
    """test estimate_computation_speed method"""

    def f(x):
        return 2 * x

    def g(x):
        return np.sin(x) * np.cos(x) ** 2

    assert misc.estimate_computation_speed(f, 1) > misc.estimate_computation_speed(g, 1)


def test_classproperty():
    """test classproperty decorator"""

    class Test:
        _value = 2

        @misc.classproperty
        def value(cls):  # @NoSelf
            return cls._value

    assert Test.value == 2


@misc.skipUnlessModule("h5py")
def test_hdf_write_attributes(tmp_path):
    """test hdf_write_attributes function"""
    import h5py

    path = tmp_path / "test_hdf_write_attributes.hdf5"

    # test normal case
    data = {"a": 3, "b": "asd"}
    with h5py.File(path, "w") as hdf_file:
        misc.hdf_write_attributes(hdf_file, data)
        data2 = {k: json.loads(v) for k, v in hdf_file.attrs.items()}

    assert data == data2
    assert data is not data2

    # test silencing of problematic items
    with h5py.File(path, "w") as hdf_file:
        misc.hdf_write_attributes(hdf_file, {"a": 1, "b": object()})
        data2 = {k: json.loads(v) for k, v in hdf_file.attrs.items()}
    assert data2 == {"a": 1}

    # test raising problematic items
    with h5py.File(path, "w") as hdf_file:
        with pytest.raises(TypeError):
            misc.hdf_write_attributes(
                hdf_file, {"a": object()}, raise_serialization_error=True
            )



"""
Package that contains base classes for solvers.

Beside the abstract base class defining the interfaces, we also provide
:class:`AdaptiveSolverBase`, which contains methods for implementing adaptive solvers.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from __future__ import annotations

import logging
import warnings
from abc import ABCMeta
from inspect import isabstract
from typing import Any, Callable, Dict, List, Optional, Tuple, Type  # @UnusedImport

import numba as nb
import numpy as np
from numba.extending import register_jitable

from ..fields.base import FieldBase
from ..pdes.base import PDEBase
from ..tools.math import OnlineStatistics
from ..tools.misc import classproperty
from ..tools.numba import is_jitted, jit
from ..tools.typing import BackendType


class SolverBase(metaclass=ABCMeta):
    """base class for solvers"""

    dt_default: float = 1e-3
    """float: default time step used if no time step was specified"""

    _modify_state_after_step: bool = True
    """bool: flag choosing whether the `modify_after_step` hook of the PDE is called"""

    _subclasses: Dict[str, Type[SolverBase]] = {}
    """dict: dictionary of all inheriting classes"""

    def __init__(self, pde: PDEBase, *, backend: BackendType = "auto"):
        """
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The partial differential equation that should be solved
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.
        """
        self.pde = pde
        self.backend = backend
        self.info: Dict[str, Any] = {"class": self.__class__.__name__}
        if self.pde:
            self.info["pde_class"] = self.pde.__class__.__name__
        self._logger = logging.getLogger(self.__class__.__name__)

    def __init_subclass__(cls, **kwargs):  # @NoSelf
        """register all subclassess to reconstruct them later"""
        super().__init_subclass__(**kwargs)
        if not isabstract(cls):
            if cls.__name__ in cls._subclasses:
                warnings.warn(f"Redefining class {cls.__name__}")
            cls._subclasses[cls.__name__] = cls
        if hasattr(cls, "name") and cls.name:
            if cls.name in cls._subclasses:
                logging.warning(f"Solver with name {cls.name} is already registered")
            cls._subclasses[cls.name] = cls

    @classmethod
    def from_name(cls, name: str, pde: PDEBase, **kwargs) -> SolverBase:
        r"""create solver class based on its name

        Solver classes are automatically registered when they inherit from
        :class:`SolverBase`. Note that this also requires that the respective python
        module containing the solver has been loaded before it is attempted to be used.

        Args:
            name (str):
                The name of the solver to construct
            pde (:class:`~pde.pdes.base.PDEBase`):
                The partial differential equation that should be solved
            \**kwargs:
                Additional arguments for the constructor of the solver

        Returns:
            An instance of a subclass of :class:`SolverBase`
        """
        try:
            # obtain the solver class associated with `name`
            solver_class = cls._subclasses[name]
        except KeyError:
            # solver was not registered
            solvers = (
                f"'{solver}'"
                for solver in sorted(cls._subclasses.keys())
                if not solver.endswith("Solver")
            )
            raise ValueError(
                f"Unknown solver method '{name}'. Registered solvers are "
                + ", ".join(solvers)
            )

        return solver_class(pde, **kwargs)

    @classproperty
    def registered_solvers(cls) -> List[str]:  # @NoSelf
        """list of str: the names of the registered solvers"""
        return list(sorted(cls._subclasses.keys()))

    @property
    def _compiled(self) -> bool:
        """bool: indicates whether functions need to be compiled"""
        return self.backend == "numba" and not nb.config.DISABLE_JIT

    def _make_modify_after_step(
        self, state: FieldBase
    ) -> Callable[[np.ndarray], float]:
        """create a function that modifies a state after each step

        A noop function will be returned if `_modify_state_after_step` is `False`,

        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted.
        """
        if self._modify_state_after_step:
            modify_after_step = jit(self.pde.make_modify_after_step(state))

        else:

            def modify_after_step(state_data: np.ndarray) -> float:
                return 0

        if self._compiled:
            sig_modify = (nb.typeof(state.data),)
            modify_after_step = jit(sig_modify)(modify_after_step)

        return modify_after_step  # type: ignore

    def _make_pde_rhs(
        self, state: FieldBase, backend: BackendType = "auto"
    ) -> Callable[[np.ndarray, float], np.ndarray]:
        """obtain a function for evaluating the right hand side

        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted.
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.

        Raises:
            RuntimeError: when a stochastic partial differential equation is encountered
            but `allow_stochastic == False`.

        Returns:
            A function that is called with data given by a :class:`~numpy.ndarray` and a
            time. The function returns the deterministic evolution rate and (if
            applicable) a realization of the associated noise.
        """
        if getattr(self.pde, "is_sde"):
            raise RuntimeError(
                f"Cannot create a deterministic stepper for a stochastic equation"
            )

        rhs = self.pde.make_pde_rhs(state, backend=backend)  # type: ignore

        if hasattr(rhs, "_backend"):
            self.info["backend"] = rhs._backend
        elif is_jitted(rhs):
            self.info["backend"] = "numba"
        else:
            self.info["backend"] = "undetermined"

        return rhs

    def _make_sde_rhs(
        self, state: FieldBase, backend: str = "auto"
    ) -> Callable[[np.ndarray, float], Tuple[np.ndarray, np.ndarray]]:
        """obtain a function for evaluating the right hand side

        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted.
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.

        Raises:
            RuntimeError: when a stochastic partial differential equation is encountered
            but `allow_stochastic == False`.

        Returns:
            A function that is called with data given by a :class:`~numpy.ndarray` and a
            time. The function returns the deterministic evolution rate and (if
            applicable) a realization of the associated noise.
        """
        rhs = self.pde.make_sde_rhs(state, backend=backend)  # type: ignore

        if hasattr(rhs, "_backend"):
            self.info["backend"] = rhs._backend
        elif is_jitted(rhs):
            self.info["backend"] = "numba"
        else:
            self.info["backend"] = "undetermined"

        return rhs

    def _make_single_step_fixed_dt(
        self, state: FieldBase, dt: float
    ) -> Callable[[np.ndarray, float], None]:
        """return a function doing a single step with a fixed time step

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the explicit stepping.
        """
        raise NotImplementedError("Fixed stepper has not been defined")

    def _make_fixed_stepper(
        self, state: FieldBase, dt: float
    ) -> Callable[[np.ndarray, float, int], Tuple[float, float]]:
        """return a stepper function using an explicit scheme with fixed time steps

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step of the explicit stepping.
        """
        single_step = self._make_single_step_fixed_dt(state, dt)
        modify_state_after_step = self._modify_state_after_step
        modify_after_step = self._make_modify_after_step(state)

        if self._compiled:
            sig_single_step = (nb.typeof(state.data), nb.double)
            single_step = jit(sig_single_step)(single_step)

        def fixed_stepper(
            state_data: np.ndarray, t_start: float, steps: int
        ) -> Tuple[float, float]:
            """perform `steps` steps with fixed time steps"""
            modifications = 0.0
            for i in range(steps):
                # calculate the right hand side
                t = t_start + i * dt
                single_step(state_data, t)
                if modify_state_after_step:
                    modifications += modify_after_step(state_data)

            return t + dt, modifications

        if self._compiled:
            sig_fixed = (nb.typeof(state.data), nb.double, nb.int_)
            fixed_stepper = jit(sig_fixed)(fixed_stepper)

        return fixed_stepper

    def make_stepper(
        self, state: FieldBase, dt: Optional[float] = None
    ) -> Callable[[FieldBase, float, float], float]:
        """return a stepper function using an explicit scheme

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step used (Uses :attr:`SolverBase.dt_default` if `None`)

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, t_end: float)`
        """
        # support `None` as a default value, so the controller can signal that
        # the solver should use a default time step
        if dt is None:
            dt = self.dt_default
            self._logger.warning(
                "Explicit stepper with a fixed time step did not receive any "
                f"initial value for `dt`. Using dt={dt}, but specifying a value or "
                "enabling adaptive stepping is advisable."
            )
        dt_float = float(dt)  # explicit casting to help type checking

        self.info["dt"] = dt_float
        self.info["steps"] = 0
        self.info["state_modifications"] = 0.0
        self.info["stochastic"] = getattr(self.pde, "is_sde", False)
        # we don't access self.pde directly since we might want to reuse the solver
        # infrastructure for more general cases where a PDE is not defined

        # create stepper with fixed steps
        fixed_stepper = self._make_fixed_stepper(state, dt_float)

        def wrapped_stepper(state: FieldBase, t_start: float, t_end: float) -> float:
            """advance `state` from `t_start` to `t_end` using fixed steps"""
            # calculate number of steps (which is at least 1)
            steps = max(1, int(np.ceil((t_end - t_start) / dt_float)))
            t_last, modifications = fixed_stepper(state.data, t_start, steps)
            self.info["steps"] += steps
            self.info["state_modifications"] += modifications
            return t_last

        return wrapped_stepper


class AdaptiveSolverBase(SolverBase):
    """base class for adaptive time steppers"""

    dt_min: float = 1e-10
    """float: minimal time step that the adaptive solver will use"""
    dt_max: float = 1e10
    """float: maximal time step that the adaptive solver will use"""

    def __init__(
        self,
        pde: PDEBase,
        *,
        backend: BackendType = "auto",
        adaptive: bool = True,
        tolerance: float = 1e-4,
    ):
        """
        Args:
            pde (:class:`~pde.pdes.base.PDEBase`):
                The instance describing the pde that needs to be solved
            backend (str):
                Determines how the function is created. Accepted  values are 'numpy` and
                'numba'. Alternatively, 'auto' lets the code decide for the most optimal
                backend.
            adaptive (bool):
                When enabled, the time step is adjusted during the simulation using the
                error tolerance set with `tolerance`.
            tolerance (float):
                The error tolerance used in adaptive time stepping. This is used in
                adaptive time stepping to choose a time step which is small enough so
                the truncation error of a single step is below `tolerance`.
        """
        super().__init__(pde, backend=backend)
        self.adaptive = adaptive
        self.tolerance = tolerance

    def _make_error_synchronizer(self) -> Callable[[float], float]:
        """return helper function that synchronizes errors between multiple processes"""

        @register_jitable
        def synchronize_errors(error: float) -> float:
            return error

        return synchronize_errors  # type: ignore

    def _make_dt_adjuster(self) -> Callable[[float, float], float]:
        """return a function that can be used to adjust time steps"""
        dt_min = self.dt_min
        dt_min_nan_err = f"Encountered NaN even though dt < {dt_min}"
        dt_min_err = f"Time step below {dt_min}"
        dt_max = self.dt_max

        def adjust_dt(dt: float, error_rel: float) -> float:
            """helper function that adjust the time step

            The goal is to keep the relative error `error_rel` close to 1.

            Args:
                dt (float): Current time step
                error_rel (float): Current (normalized) error estimate

            Returns:
                float: Time step of the next iteration
            """
            # adjust the time step
            if error_rel < 0.00057665:
                # error was very small => maximal increase in dt
                # The constant on the right hand side of the comparison is chosen to
                # agree with the equation for adjusting dt below
                dt *= 4.0
            elif np.isnan(error_rel):
                # state contained NaN => decrease time step strongly
                dt *= 0.25
            else:
                # otherwise, adjust time step according to error
                dt *= max(0.9 * error_rel**-0.2, 0.1)

            # limit time step to permissible bracket
            if dt > dt_max:
                dt = dt_max
            elif dt < dt_min:
                if np.isnan(error_rel):
                    raise RuntimeError(dt_min_nan_err)
                else:
                    raise RuntimeError(dt_min_err)

            return dt

        if self._compiled:
            adjust_dt = jit((nb.double, nb.double))(adjust_dt)

        return adjust_dt

    def _make_single_step_variable_dt(
        self, state: FieldBase
    ) -> Callable[[np.ndarray, float, float], np.ndarray]:
        """return a function doing a single step with a variable time step

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is
            `(state: numpy.ndarray, t_start: float, t_end: float)`
        """
        rhs_pde = self._make_pde_rhs(state, backend=self.backend)

        def single_step(state_data: np.ndarray, t: float, dt: float) -> np.ndarray:
            """basic implementation of Euler scheme"""
            return state_data + dt * rhs_pde(state_data, t)  # type: ignore

        return single_step

    def _make_single_step_error_estimate(
        self, state: FieldBase
    ) -> Callable[[np.ndarray, float, float], Tuple[np.ndarray, float]]:
        """make a stepper that also estimates the error

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
        """
        if getattr(self.pde, "is_sde"):
            raise RuntimeError("Cannot use adaptive stepper with stochastic equation")

        single_step = self._make_single_step_variable_dt(state)
        if self._compiled:
            sig_single_step = (nb.typeof(state.data), nb.double, nb.double)
            single_step = jit(sig_single_step)(single_step)

        def single_step_error_estimate(
            state_data: np.ndarray, t: float, dt: float
        ) -> Tuple[np.ndarray, float]:
            """basic stepper to estimate error"""
            # single step with dt
            k1 = single_step(state_data, t, dt)

            # double step with half the time step
            k2a = single_step(state_data, t, 0.5 * dt)
            k2 = single_step(k2a, t + 0.5 * dt, 0.5 * dt)

            # calculate maximal error
            error = np.abs(k1 - k2).max()

            return k2, error

        return single_step_error_estimate

    def _make_adaptive_stepper(
        self, state: FieldBase
    ) -> Callable[
        [np.ndarray, float, float, float, Optional[OnlineStatistics]],
        Tuple[float, float, int, float],
    ]:
        """make an adaptive Euler stepper

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, t_end: float)`
        """
        # obtain functions determining how the PDE is evolved
        single_step_error = self._make_single_step_error_estimate(state)
        modify_after_step = self._make_modify_after_step(state)
        modify_state_after_step = self._modify_state_after_step
        sync_errors = self._make_error_synchronizer()

        # obtain auxiliary functions
        adjust_dt = self._make_dt_adjuster()
        tolerance = self.tolerance
        dt_min = self.dt_min

        if self._compiled:
            # compile paired stepper
            sig_stepper = (nb.typeof(state.data), nb.double, nb.double)
            single_step_error = jit(sig_stepper)(single_step_error)

        def adaptive_stepper(
            state_data: np.ndarray,
            t_start: float,
            t_end: float,
            dt_init: float,
            dt_stats: Optional[OnlineStatistics] = None,
        ) -> Tuple[float, float, int, float]:
            """adaptive stepper that advances the state in time"""
            modifications = 0.0
            dt_opt = dt_init
            t = t_start
            steps = 0
            while True:
                # use a smaller (but not too small) time step if close to t_end
                dt_step = max(min(dt_opt, t_end - t), dt_min)

                # two different steppings to estimate errors
                new_state, error = single_step_error(state_data, t, dt_step)

                error_rel = error / tolerance  # normalize error to given tolerance
                # synchronize the error between all processes (if necessary)
                error_rel = sync_errors(error_rel)

                # do the step if the error is sufficiently small
                if error_rel <= 1:
                    steps += 1
                    t += dt_step
                    state_data[...] = new_state
                    if modify_state_after_step:
                        modifications += modify_after_step(state_data)
                    if dt_stats is not None:
                        dt_stats.add(dt_step)

                if t < t_end:
                    # adjust the time step and continue
                    dt_opt = adjust_dt(dt_step, error_rel)
                else:
                    break  # return to the controller

            return t, dt_opt, steps, modifications

        if self._compiled:
            # compile inner stepper
            sig_adaptive = (
                nb.typeof(state.data),
                nb.double,
                nb.double,
                nb.double,
                nb.typeof(self.info["dt_statistics"]),
            )
            adaptive_stepper = jit(sig_adaptive)(adaptive_stepper)

        self._logger.info(f"Initialized adaptive stepper")
        return adaptive_stepper

    def make_stepper(
        self, state: FieldBase, dt: Optional[float] = None
    ) -> Callable[[FieldBase, float, float], float]:
        """return a stepper function using an explicit scheme

        Args:
            state (:class:`~pde.fields.base.FieldBase`):
                An example for the state from which the grid and other information can
                be extracted
            dt (float):
                Time step used (Uses :attr:`SolverBase.dt_default` if `None`). This sets
                the initial time step for adaptive solvers.

        Returns:
            Function that can be called to advance the `state` from time `t_start` to
            time `t_end`. The function call signature is `(state: numpy.ndarray,
            t_start: float, t_end: float)`
        """
        if not self.adaptive:
            # create stepper with fixed steps
            return super().make_stepper(state, dt)

        # support `None` as a default value, so the controller can signal that
        # the solver should use a default time step
        if dt is None:
            dt_float = self.dt_default
        else:
            dt_float = float(dt)  # explicit casting to help type checking

        self.info["dt"] = dt_float
        self.info["dt_adaptive"] = True
        self.info["steps"] = 0
        self.info["stochastic"] = getattr(self.pde, "is_sde", False)
        self.info["state_modifications"] = 0.0

        # create stepper with adaptive steps
        self.info["dt_statistics"] = OnlineStatistics()
        adaptive_stepper = self._make_adaptive_stepper(state)

        def wrapped_stepper(state: FieldBase, t_start: float, t_end: float) -> float:
            """advance `state` from `t_start` to `t_end` using adaptive steps"""
            nonlocal dt_float  # `dt_float` stores value for the next call

            t_last, dt_float, steps, modifications = adaptive_stepper(
                state.data, t_start, t_end, dt_float, self.info["dt_statistics"]
            )
            self.info["steps"] += steps
            self.info["state_modifications"] += modifications
            return t_last

        return wrapped_stepper



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import itertools

import pytest

from pde import UnitGrid
from pde.grids.boundaries.axis import BoundaryPair, get_boundary_axis
from pde.grids.boundaries.local import BCBase


def test_boundary_pair():
    """test setting boundary conditions for whole axis"""
    g = UnitGrid([2, 3])
    b = ["value", {"type": "derivative", "value": 1}]
    for bl, bh in itertools.product(b, b):
        bc = BoundaryPair.from_data(g, 0, [bl, bh])
        blo = BCBase.from_data(g, 0, upper=False, data=bl)
        bho = BCBase.from_data(g, 0, upper=True, data=bh)

        assert bc.low == blo
        assert bc.high == bho
        assert bc == BoundaryPair(blo, bho)
        if bl == bh:
            assert bc == BoundaryPair.from_data(g, 0, bl)
        assert list(bc) == [blo, bho]
        assert isinstance(str(bc), str)
        assert isinstance(repr(bc), str)

        bc.check_value_rank(0)
        with pytest.raises(RuntimeError):
            bc.check_value_rank(1)

    data = {"low": {"value": 1}, "high": {"derivative": 2}}
    bc1 = BoundaryPair.from_data(g, 0, data)
    bc2 = BoundaryPair.from_data(g, 0, data)
    assert bc1 == bc2 and bc1 is not bc2
    bc2 = BoundaryPair.from_data(g, 1, data)
    assert bc1 != bc2 and bc1 is not bc2

    # miscellaneous methods
    data = {"low": {"value": 0}, "high": {"derivative": 0}}
    bc1 = BoundaryPair.from_data(g, 0, data)
    b_lo, b_hi = bc1
    assert b_lo == BCBase.from_data(g, 0, False, {"value": 0})
    assert b_hi == BCBase.from_data(g, 0, True, {"derivative": 0})
    assert b_lo is bc1[0]
    assert b_lo is bc1[False]
    assert b_hi is bc1[1]
    assert b_hi is bc1[True]


def test_get_axis_boundaries():
    """test setting boundary conditions including periodic ones"""
    for data in ["value", "derivative", "periodic", "anti-periodic"]:
        g = UnitGrid([2], periodic=("periodic" in data))
        b = get_boundary_axis(g, 0, data)
        assert str(b) == '"' + data + '"'
        b1, b2 = b.get_mathematical_representation("field")
        assert "field" in b1 and "field" in b2

        if "periodic" in data:
            assert b.periodic
            assert len(list(b)) == 2
            assert b.flip_sign == (data == "anti-periodic")
        else:
            assert not b.periodic
            assert len(list(b)) == 2



"""
Module defining classes for tracking results from simulations.

The trackers defined in this module are:

.. autosummary::
   :nosignatures:

   CallbackTracker
   ProgressTracker
   PrintTracker
   PlotTracker
   LivePlotTracker
   DataTracker
   SteadyStateTracker
   RuntimeTracker
   ConsistencyTracker
   MaterialConservationTracker

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import inspect
import math
import os.path
import sys
import time
from datetime import timedelta
from pathlib import Path
from typing import List  # @UnusedImport
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, Optional, Union

import numpy as np

from ..fields import FieldCollection
from ..fields.base import FieldBase
from ..tools.docstrings import fill_in_docstring
from ..tools.misc import module_available
from ..tools.output import get_progress_bar_class
from ..tools.parse_duration import parse_duration
from ..tools.typing import Real
from .base import FinishedSimulation, InfoDict, TrackerBase
from .interrupts import IntervalData, RealtimeInterrupts

if TYPE_CHECKING:
    import pandas  # @UnusedImport

    from ..visualization.movies import Movie  # @UnusedImport


class CallbackTracker(TrackerBase):
    """Tracker calling a function periodically

    Example:
        The callback tracker can be used to check for conditions during the simulation:

        .. code-block:: python

            def check_simulation(state, time):
                if state.integral < 0:
                    raise StopIteration

            tracker = CallbackTracker(check_simulation, interval="0:10")

        Adding :code:`tracker` to the simulation will perform a check every 10 real time
        seconds. If the integral of the entire state falls below zero, the simulation
        will be aborted.
    """

    @fill_in_docstring
    def __init__(self, func: Callable, interval: IntervalData = 1):
        """
        Args:
            func: The function to call periodically. The function signature should be
                `(state)` or `(state, time)`, where `state` contains the current state
                as an instance of :class:`~pde.fields.base.FieldBase` and `time` is a
                float value indicating the current time. Note that only a view of the
                state is supplied, implying that a copy needs to be made if the data
                should be stored. The function can thus adjust the state by modifying it
                in-place and it can even interrupt the simulation by raising the special
                exception :class:`StopIteration`.
            interval:
                {ARG_TRACKER_INTERVAL}
        """
        super().__init__(interval=interval)
        self._callback = func
        self._num_args = len(inspect.signature(func).parameters)
        if not 0 < self._num_args < 3:
            raise ValueError(
                "`func` must be a function accepting one or two arguments, not "
                f"{self._num_args}"
            )

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if self._num_args == 1:
            self._callback(field)
        else:
            self._callback(field, t)


class ProgressTracker(TrackerBase):
    """Tracker showing the progress of the simulation"""

    name = "progress"

    @fill_in_docstring
    def __init__(
        self,
        interval: Optional[IntervalData] = None,
        *,
        fancy: bool = True,
        ndigits: int = 5,
        leave: bool = True,
    ):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
                The default value `None` updates the progress bar approximately every
                (real) second.
            fancy (bool):
                Flag determining whether a fancy progress bar should be used in jupyter
                notebooks (if :mod:`ipywidgets` is installed)
            ndigits (int):
                The number of digits after the decimal point that are shown maximally.
            leave (bool):
                Whether to leave the progress bar after the simulation has finished
                (default: True)
        """
        if interval is None:
            interval = RealtimeInterrupts(duration=1)  # print every second by default

        super().__init__(interval=interval)
        self.fancy = fancy
        self.ndigits = ndigits
        self.leave = leave

    def initialize(self, field: FieldBase, info: Optional[InfoDict] = None) -> float:
        """initialize the tracker with information about the simulation

        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        result = super().initialize(field, info)

        # get solver information
        controller_info = {} if info is None else info.get("controller", {})

        # initialize the progress bar
        pb_cls = get_progress_bar_class(self.fancy)
        self.progress_bar = pb_cls(
            total=controller_info.get("t_end"),
            initial=controller_info.get("t_start", 0),
            leave=self.leave,
        )
        self.progress_bar.set_description("Initializing")

        return result

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        # show an update
        if self.progress_bar.total:
            t_new = min(t, self.progress_bar.total)
        else:
            t_new = t
        self.progress_bar.n = round(t_new, self.ndigits)
        self.progress_bar.set_description("")

    def finalize(self, info: Optional[InfoDict] = None) -> None:
        """finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation
        """
        super().finalize(info)
        self.progress_bar.set_description("")

        # limit progress bar to 100%
        controller_info = {} if info is None else info.get("controller", {})
        t_final = controller_info.get("t_final", -math.inf)
        t_end = controller_info.get("t_end", -math.inf)
        if t_final >= t_end and self.progress_bar.total:
            self.progress_bar.n = self.progress_bar.total
            self.progress_bar.refresh()

        if (
            controller_info.get("successful", False)
            and self.leave
            and hasattr(self.progress_bar, "sp")
        ):
            # show progress bar in green if simulation was successful. We
            # need to overwrite the default behavior (and disable the
            # progress bar) since reaching steady state means the simulation
            # was successful even though it did not reach t_final
            try:
                self.progress_bar.sp(bar_style="success")
            except TypeError:
                self.progress_bar.close()
            else:
                self.progress_bar.disable = True
        else:
            self.progress_bar.close()

    def __del__(self):
        if hasattr(self, "progress_bar") and not self.progress_bar.disable:
            self.progress_bar.close()


class PrintTracker(TrackerBase):
    """Tracker printing data to a stream (default: stdout)"""

    name = "print"

    @fill_in_docstring
    def __init__(self, interval: IntervalData = 1, stream: IO[str] = sys.stdout):
        """

        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            stream:
                The stream used for printing
        """
        super().__init__(interval=interval)
        self.stream = stream

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        data = f"c={field.data.mean():.3g}±{field.data.std():.3g}"

        self.stream.write(f"t={t:g}, {data}\n")
        self.stream.flush()


class PlotTracker(TrackerBase):
    """Tracker plotting data on screen, to files, or writes a movie

    This tracker can be used to create movies from simulations or to simply update a
    single image file on the fly (i.e. to monitor simulations running on a cluster). The
    default values of this tracker are chosen with regular output to a file in mind.

    Example:
        To create a movie while running the simulation, you can use

        .. code-block:: python

            movie_tracker = PlotTracker(interval=10, movie="my_movie.mp4")
            eq.solve(..., tracker=movie_tracker)

        This will create the file `my_movie.mp4` during the simulation. Note that you
        can display the frames interactively by setting :code:`show=True`.
    """

    @fill_in_docstring
    def __init__(
        self,
        interval: IntervalData = 1,
        *,
        title: Union[str, Callable] = "Time: {time:g}",
        output_file: Optional[str] = None,
        movie: Union[str, Path, "Movie", None] = None,
        show: Optional[bool] = None,
        tight_layout: bool = False,
        max_fps: float = math.inf,
        plot_args: Optional[Dict[str, Any]] = None,
    ):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            title (str or callable):
                Title text of the figure. If this is a string, it is shown with a
                potential placeholder named `time` being replaced by the current
                simulation time. Conversely, if `title` is a function, it is called with
                the current state and the time as arguments. This function is expected
                to return a string.
            output_file (str, optional):
                Specifies a single image file, which is updated periodically, so that
                the progress can be monitored (e.g. on a compute cluster)
            movie (str or :class:`~pde.visualization.movies.Movie`):
                Create a movie. If a filename is given, all frames are written to this
                file in the format deduced from the extension after the simulation ran.
                If a :class:`~pde.visualization.movies.Movie` is supplied, frames are
                appended to the instance.
            show (bool, optional):
                Determines whether the plot is shown while the simulation is running. If
                set to `None`, the images are only shown if neither `output_file` nor
                `movie` is set, otherwise they are kept hidden. Note that showing the
                plot can slow down a simulation severely.
            tight_layout (bool):
                Determines whether :func:`~matplotlib.pyplot.tight_layout` is used.
            max_fps (float):
                Determines the maximal rate (frames per second) at which the plots are
                updated in real time during the simulation. Some plots are skipped if
                the tracker receives data at a higher rate. A larger value (e.g.,
                `math.inf`) can be used to ensure every frame is drawn, which might
                penalizes the overall performance.
            plot_args (dict):
                Extra arguments supplied to the plot call. For example, this can be used
                to specify axes ranges when a single panel is shown. For instance, the
                value :code:`{'ax_style': {'ylim': (0, 1)}}` enforces the y-axis to lie
                between 0 and 1.

        Note:
            If an instance of :class:`~pde.visualization.movies.Movie` is given as the
            `movie` argument, it can happen that the movie is not written to the file
            when the simulation ends. This is because, the movie could still be extended
            by appending frames. To write the movie to a file call its
            :meth:`~pde.visualization.movies.Movie.save` method. Beside adding frames
            before and after the simulation, an explicit movie object can also be used
            to adjust the output. For instance, the following example code creates a
            movie with a framerate of 15, a resolution of 200 dpi, and a bitrate of 6000
            kilobits per second:

            .. code-block:: python

                movie = Movie("movie.mp4", framerate=15, dpi=200, bitrate=6000)
                eq.solve(..., tracker=PlotTracker(1, movie=movie))
                movie.save()
        """
        from ..visualization.movies import Movie  # @Reimport

        # initialize the tracker
        super().__init__(interval=interval)
        self.title = title
        self.output_file = output_file
        self.tight_layout = tight_layout
        self.max_fps = max_fps

        self.plot_args = {} if plot_args is None else plot_args.copy()
        # make sure the plot is only create and not shown since the context
        # handles showing the plot itself
        self.plot_args["action"] = "none"

        # initialize the movie class
        if movie is None:
            self.movie: Optional[Movie] = None
            self._save_movie = False

        elif isinstance(movie, Movie):
            self.movie = movie
            self._save_movie = False

        elif isinstance(movie, (str, Path)):
            self.movie = Movie(filename=str(movie))
            self._save_movie = True

        else:
            raise TypeError(f"Unknown type of `movie`: {movie.__class__.__name__}")

        # determine whether to show the images interactively
        self._write_images = self._save_movie or self.output_file
        if show is None:
            self.show = not self._write_images
        else:
            self.show = show

    def initialize(self, state: FieldBase, info: Optional[InfoDict] = None) -> float:
        """initialize the tracker with information about the simulation

        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        # initialize the plotting context
        import matplotlib.pyplot as plt

        from ..tools.plotting import get_plotting_context

        self._context = get_plotting_context(title="Initializing...", show=self.show)

        # do the actual plotting
        with self._context:
            self._plot_reference = state.plot(**self.plot_args)
            if self.tight_layout:
                plt.gcf().tight_layout()

        if self._context.supports_update:
            # the context supports reusing figures
            if hasattr(state.plot, "update_method"):
                # the plotting method supports updating the plot
                if state.plot.update_method is None:
                    if state.plot.mpl_class == "axes":  # type: ignore
                        self._update_method = "update_ax"
                    elif state.plot.mpl_class == "figure":  # type: ignore
                        self._update_method = "update_fig"
                    else:
                        mpl_class = state.plot.mpl_class  # type: ignore
                        raise RuntimeError(
                            f"Unknown mpl_class on plot method: {mpl_class}"
                        )
                else:
                    self._update_method = "update_data"
            else:
                raise RuntimeError(
                    "PlotTracker does not  work since the state of type "
                    f"{state.__class__.__name__} does not use the plot protocol of "
                    "`pde.tools.plotting`."
                )
        else:
            self._update_method = "replot"

        self._logger.info(f'Update method: "{self._update_method}"')
        self._last_update = time.monotonic()
        return super().initialize(state, info=info)

    def handle(self, state: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            state (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        import matplotlib.pyplot as plt

        if not self._write_images:
            # check whether we can skip this image
            time_passed = time.monotonic() - self._last_update
            if time_passed < 1 / self.max_fps:
                return  # we just recently updated the image

        if callable(self.title):
            self._context.title = str(self.title(state, t))
        else:
            self._context.title = self.title.format(time=t)

        # update the plot in the correct plotting context
        with self._context:
            if self._update_method == "update_data":
                # the state supports updating the plot data
                update_func = getattr(state, state.plot.update_method)  # type: ignore
                update_func(self._plot_reference)

            elif self._update_method == "update_ax":
                fig = self._context.fig
                fig.clf()  # type: ignore
                ax = fig.add_subplot(1, 1, 1)  # type: ignore
                state.plot(ax=ax, **self.plot_args)

            elif self._update_method == "update_fig":
                fig = self._context.fig
                fig.clf()  # type: ignore
                state.plot(fig=fig, **self.plot_args)
                if self.tight_layout:
                    plt.gcf().tight_layout()

            elif self._update_method == "replot":
                state.plot(**self.plot_args)
                if self.tight_layout:
                    plt.gcf().tight_layout()

            else:
                raise RuntimeError(f"Unknown update method `{self._update_method}`")

        if self.output_file and self._context.fig is not None:
            self._context.fig.savefig(self.output_file)
        if self.movie:
            self.movie.add_figure(self._context.fig)

        self._last_update = time.monotonic()

    def finalize(self, info: Optional[InfoDict] = None) -> None:
        """finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation
        """
        super().finalize(info)
        if self._save_movie:
            # write out movie file
            self.movie.save()  # type: ignore
            # end recording the movie (e.g. delete temporary files)
            self.movie._end()  # type: ignore

        if not self.show:
            self._context.close()


class LivePlotTracker(PlotTracker):
    """PlotTracker with defaults for live plotting

    The only difference to :class:`PlotTracker` are the changed default values, where
    output is by default shown on screen and the `interval` is set something more
    suitable for interactive plotting. In particular, this tracker can be enabled by
    simply listing 'plot' as a tracker.
    """

    name = "plot"

    @fill_in_docstring
    def __init__(
        self,
        interval: IntervalData = "0:03",
        *,
        show: bool = True,
        max_fps: float = 2,
        **kwargs,
    ):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            title (str):
                Text to show in the title. The current time point will be
                appended to this text, so include a space for optimal results.
            output_file (str, optional):
                Specifies a single image file, which is updated periodically, so
                that the progress can be monitored (e.g. on a compute cluster)
            output_folder (str, optional):
                Specifies a folder to which all images are written. The files
                will have names with increasing numbers.
            movie_file (str, optional):
                Specifies a filename to which a movie of all the frames is
                written after the simulation.
            show (bool, optional):
                Determines whether the plot is shown while the simulation is
                running. If `False`, the files are created in the background.
                This option can slow down a simulation severely.
            max_fps (float):
                Determines the maximal rate (frames per second) at which the plots are
                updated. Some plots are skipped if the tracker receives data at a higher
                rate. A larger value (e.g., `math.inf`) can be used to ensure every
                frame is drawn, which might penalizes the overall performance.
            plot_args (dict):
                Extra arguments supplied to the plot call. For example, this can
                be used to specify axes ranges when a single panel is shown. For
                instance, the value `{'ax_style': {'ylim': (0, 1)}}` enforces
                the y-axis to lie between 0 and 1.
        """
        super().__init__(interval=interval, show=show, max_fps=max_fps, **kwargs)


class DataTracker(CallbackTracker):
    """Tracker storing custom data obtained by calling a function

    Example:
        The data tracker can be used to gather statistics during the run

        .. code-block:: python

            def get_statistics(state, time):
                return {"mean": state.data.mean(), "variance": state.data.var()}

            data_tracker = DataTracker(get_statistics, interval=10)

        Adding :code:`data_tracker` to the simulation will gather the statistics every
        10 time units. After the simulation, the final result will be accessable via the
        :attr:`data` attribute or conveniently as a pandas from the :attr:`dataframe`
        attribute.

    Attributes:
        times (list):
            The time points at which the data is stored
        data (list):
            The actually stored data, which is a list of the objects returned by
            the callback function.
    """

    @fill_in_docstring
    def __init__(
        self, func: Callable, interval: IntervalData = 1, filename: Optional[str] = None
    ):
        """
        Args:
            func:
                The function to call periodically. The function signature
                should be `(state)` or `(state, time)`, where `state` contains
                the current state as an instance of
                :class:`~pde.fields.FieldBase` and `time` is a
                float value indicating the current time. Note that only a view
                of the state is supplied, implying that a copy needs to be made
                if the data should be stored.
                Typical return values of the function are either a single
                number, a numpy array, a list of number, or a dictionary to
                return multiple numbers with assigned labels.
            interval:
                {ARG_TRACKER_INTERVAL}
            filename (str):
                A path to a file to which the data is written at the end of the
                tracking. The data format will be determined by the extension
                of the filename. '.pickle' indicates a python pickle file
                storing a tuple `(self.times, self.data)`, whereas any other
                data format requires :mod:`pandas`.
        """
        super().__init__(func=func, interval=interval)
        self.filename = filename
        self.times: List[float] = []
        self.data: List[Any] = []

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        self.times.append(t)
        if self._num_args == 1:
            self.data.append(self._callback(field))
        else:
            self.data.append(self._callback(field, t))

    def finalize(self, info: Optional[InfoDict] = None) -> None:
        """finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation
        """
        super().finalize(info)
        if self.filename:
            self.to_file(self.filename)

    @property
    def dataframe(self) -> "pandas.DataFrame":
        """:class:`pandas.DataFrame`: the data in a dataframe

        If `func` returns a dictionary, the keys are used as column names.
        Otherwise, the returned data is enumerated starting with '0'. In any
        case the time point at which the data was recorded is stored in the
        column 'time'.
        """
        import pandas as pd

        df = pd.DataFrame(self.data)
        # insert the times and use them as an index
        df.insert(0, "time", self.times)
        return df

    def to_file(self, filename: str, **kwargs):
        r"""store data in a file

        The extension of the filename determines what format is being used. For
        instance, '.pickle' indicates a python pickle file storing a tuple
        `(self.times, self.data)`, whereas any other data format requires
        :mod:`pandas`. Supported formats include 'csv', 'json'.

        Args:
            filename (str):
                Path where the data is stored
            \**kwargs:
                Additional parameters may be supported for some formats
        """
        extension = os.path.splitext(filename)[1].lower()
        if extension == ".pickle":
            import pickle

            with open(filename, "wb") as fp:
                pickle.dump((self.times, self.data), fp, **kwargs)

        elif extension == ".csv":
            self.dataframe.to_csv(filename, **kwargs)
        elif extension == ".json":
            self.dataframe.to_json(filename, **kwargs)
        elif extension in {".xls", ".xlsx"}:
            self.dataframe.to_excel(filename, **kwargs)
        else:
            raise ValueError(f"Unsupported file extension `{extension}`")


class SteadyStateTracker(TrackerBase):
    """Tracker aborting the simulation once steady state is reached

    Steady state is obtained when the state does not change anymore, i.e., when the
    evolution rate is close to zero. If the argument `evolution_rate` is specified, it
    is used to calculate the evolution rate directly. If it is omitted, the evolution
    rate is estaimted by comparing the current state `cur` to the state `prev` at the
    previous time step. In both cases, convergence is assumed when the absolute value of
    the evolution rate falls below :code:`atol + rtol * cur` for all points. Here,
    `atol` and `rtol` denote absolute and relative tolerances, respectively.
    """

    name = "steady_state"

    progress_bar_format = (
        "Convergence: {percentage:3.0f}%|{bar}| [{elapsed}<{remaining}]"
    )
    """ determines the format of the progress bar shown when `progress = True` """

    @fill_in_docstring
    def __init__(
        self,
        interval: Optional[IntervalData] = None,
        atol: float = 1e-8,
        rtol: float = 1e-5,
        *,
        progress: bool = False,
        evolution_rate: Optional[Callable[[np.ndarray, float], np.ndarray]] = None,
    ):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
                The default value `None` checks for the steady state approximately every
                (real) second.
            atol (float):
                Absolute tolerance that must be reached to abort the simulation
            rtol (float):
                Relative tolerance that must be reached to abort the simulation
            progress (bool):
                Flag indicating whether the progress towards convergence is shown
                graphically during the simulation
            evolution_rate (callable):
                Function to evaluate the current evolution rate. If omitted, the
                evolution rate is estimate from the change in the state variable, which
                can be less accurate. A suitable form of the function is returned by
                `eq.make_pde_rhs(state)` when `eq` is the PDE class.
        """
        if interval is None:
            interval = RealtimeInterrupts(duration=1)
        super().__init__(interval=interval)
        self.atol = atol
        self.rtol = rtol
        self.evolution_rate = evolution_rate
        self.progress = progress and module_available("tqdm")

        self._progress_bar: Any = None
        self._pbar_offset: float = 0  # required for calculating progress
        self._last_data: Optional[np.ndarray] = None
        self._last_time: Optional[float] = None
        self._best_rate_max: Optional[np.ndarray] = None

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        finite = np.isfinite(field.data)  # ignore infinite and nan data

        # determine the maximal rate of change
        if self.evolution_rate is not None:
            # use the evolution_rate function to calculate the rate
            evolution_rate = self.evolution_rate(field.data, t)[finite]

        elif self._last_data is not None:
            # get evolution rate from the difference of current state to previous one
            diff = self._last_data[finite] - field.data[finite]
            evolution_rate = diff / (t - self._last_time)  # type: ignore

            # save current data for next comparison
            self._last_data[:] = field.data
            self._last_time = t

        else:
            # create storage for the data
            self._last_data = field.data.copy()
            self._last_time = t
            return  # do not output anything since we don't know `evolution_rate` yet

        # calculate the maximal deviation of the evolution rate from zero, subtracting
        # the relative tolerance with respect to the field values
        rate_abs = np.abs(evolution_rate) - self.rtol * np.abs(field.data[finite])
        rate_abs_max = np.max(rate_abs)

        # check wether the simulation has converged
        if rate_abs_max <= self.atol:
            if self.progress and self._progress_bar is not None:
                # advance progress bar to 100%
                self._progress_bar.n = self._pbar_offset - np.log10(self.atol)
                try:
                    self._progress_bar.disp(bar_style="success", check_delay=False)
                except (TypeError, AttributeError):
                    self._progress_bar.close()
            raise FinishedSimulation("Reached stationary state")

        if self.progress:
            # show progress of the convergence
            if self._best_rate_max is None:
                # initialize the progress bar
                pb_cls = get_progress_bar_class()
                self._pbar_offset = np.log10(rate_abs_max)
                self._progress_bar = pb_cls(
                    total=self._pbar_offset - np.log10(self.atol),
                    bar_format=self.progress_bar_format,
                )
                self._best_rate_max = rate_abs_max

            elif rate_abs_max < self._best_rate_max:
                # update progress bar if simulation got closer to convergence
                self._progress_bar.n = self._pbar_offset - np.log10(rate_abs_max)
                self._progress_bar.refresh()
                self._best_rate_max = rate_abs_max


class RuntimeTracker(TrackerBase):
    """Tracker interrupting the simulation once a duration has passed"""

    @fill_in_docstring
    def __init__(self, max_runtime: Union[Real, str], interval: IntervalData = 1):
        """
        Args:
            max_runtime (float or str):
                The maximal runtime of the simulation. If the runtime is exceeded, the
                simulation is interrupted. Values can be either given as a number
                (interpreted as seconds) or as a string, which is then parsed using the
                function :func:`~pde.tools.parse_duration.parse_duration`.
            interval:
                {ARG_TRACKER_INTERVAL}
        """
        super().__init__(interval=interval)

        try:
            self.max_runtime = float(max_runtime)
        except ValueError:
            td = parse_duration(str(max_runtime))
            self.max_runtime = td.total_seconds()

    def initialize(self, field: FieldBase, info: Optional[InfoDict] = None) -> float:
        """
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        self.max_time = time.monotonic() + self.max_runtime
        return super().initialize(field, info)

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if time.monotonic() > self.max_time:
            dt = timedelta(seconds=self.max_runtime)
            raise FinishedSimulation(f"Reached maximal runtime of {str(dt)}")


class ConsistencyTracker(TrackerBase):
    """Tracker interrupting the simulation when the state is not finite"""

    name = "consistency"

    @fill_in_docstring
    def __init__(self, interval: Optional[IntervalData] = None):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
                The default value `None` checks for consistency approximately every
                (real) second.
        """
        if interval is None:
            interval = RealtimeInterrupts(duration=1)
        super().__init__(interval=interval)

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if not np.all(np.isfinite(field.data)):
            raise StopIteration("Field was not finite")


class MaterialConservationTracker(TrackerBase):
    """Tracking interrupting the simulation when material conservation is broken"""

    name = "material_conservation"

    @fill_in_docstring
    def __init__(
        self, interval: IntervalData = 1, atol: float = 1e-4, rtol: float = 1e-4
    ):
        """
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            atol (float):
                Absolute tolerance for amount deviations
            rtol (float):
                Relative tolerance for amount deviations
        """
        super().__init__(interval=interval)
        self.atol = atol
        self.rtol = rtol

    def initialize(self, field: FieldBase, info: Optional[InfoDict] = None) -> float:
        """
        Args:
            field (:class:`~pde.fields.base.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation

        Returns:
            float: The first time the tracker needs to handle data
        """
        if isinstance(field, FieldCollection):
            self._reference = np.array([f.magnitude for f in field])
        else:
            self._reference = field.magnitude  # type: ignore

        return super().initialize(field, info)

    def handle(self, field: FieldBase, t: float) -> None:
        """handle data supplied to this tracker

        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float):
                The associated time
        """
        if isinstance(field, FieldCollection):
            mags = np.array([f.magnitude for f in field])
        else:
            mags = field.magnitude  # type: ignore

        c = np.isclose(mags, self._reference, rtol=self.rtol, atol=self.atol)
        if not np.all(c):
            if isinstance(field, FieldCollection):
                msg = f"Material of field {np.flatnonzero(~c)} is not conserved"
            else:
                msg = f"Material is not conserved"
            raise StopIteration(msg)


__all__ = [
    "CallbackTracker",
    "ProgressTracker",
    "PrintTracker",
    "PlotTracker",
    "LivePlotTracker",
    "DataTracker",
    "SteadyStateTracker",
    "RuntimeTracker",
    "ConsistencyTracker",
    "MaterialConservationTracker",
]



#!/usr/bin/env python3

import argparse
import os
import subprocess as sp
import sys
from pathlib import Path
from typing import Sequence, Union

PACKAGE = "pde"  # name of the package that needs to be tested
PACKAGE_PATH = Path(__file__).resolve().parents[1]  # base path of the package


def _most_severe_exit_code(retcodes: Sequence[int]) -> int:
    """returns the most severe exit code of a given list

    Args:
        retcodes (list): A list of return codes

    Returns:
        int: the exit code that is most severe
    """
    if all(retcode == 0 for retcode in retcodes):
        return 0
    else:
        return max(retcodes, key=lambda retcode: abs(retcode))


def show_config():
    """show package configuration"""
    from importlib.machinery import SourceFileLoader

    # imports the package from the package path
    mod = SourceFileLoader(PACKAGE, f"{PACKAGE_PATH/PACKAGE}/__init__.py").load_module()
    # obtain the package environment
    env = mod.environment()

    print(f"{'='*33} CONFIGURATION {'='*32}")
    for category, data in env.items():
        if hasattr(data, "items"):
            print(f"\n{category}:")
            for key, value in data.items():
                print(f"    {key}: {value}")
        else:
            data_formatted = data.replace("\n", "\n    ")
            print(f"{category}: {data_formatted}")
    print("=" * 80)


def test_codestyle(*, verbose: bool = True) -> int:
    """run the codestyle tests

    Args:
        verbose (bool): Whether to do extra output

    Returns:
        int: The most severe exit code
    """
    retcodes = []

    for folder in [PACKAGE, "examples"]:
        if verbose:
            print(f"Checking codestyle in folder {folder}...")
        path = PACKAGE_PATH / folder

        # format imports
        result = sp.run(["isort", "--diff", path])
        retcodes.append(result.returncode)
        # format rest
        result = sp.run(["black", "--check", path])
        retcodes.append(result.returncode)

    return _most_severe_exit_code(retcodes)


def test_types(*, report: bool = False, verbose: bool = True) -> int:
    """run mypy to check the types of the python code

    Args:
        report (bool): Whether to write a report
        verbose (bool): Whether to do extra output

    Returns:
        int: The return code indicating success or failure
    """
    if verbose:
        print(f"Checking types in the {PACKAGE} package...")

    args = [
        sys.executable,
        "-m",
        "mypy",
        "--config-file",
        PACKAGE_PATH / "pyproject.toml",
    ]

    if report:
        folder = PACKAGE_PATH / "tests" / "mypy-report"
        if verbose:
            print(f"Writing report to `{folder}`")
        args.extend(
            ["--no-incremental", "--linecount-report", folder, "--html-report", folder]
        )
    else:
        # do not create a report
        args.append("--pretty")

    args.extend(["--package", PACKAGE])

    return sp.run(args, cwd=PACKAGE_PATH).returncode


def run_unit_tests(
    *,
    runslow: bool = False,
    runinteractive: bool = False,
    use_mpi: bool = False,
    num_cores: Union[str, int] = 1,
    coverage: bool = False,
    nojit: bool = False,
    early: bool = False,
    pattern: str = None,
) -> int:
    """run the unit tests

    Args:
        runslow (bool): Whether to run the slow tests
        runinteractive (bool): Whether to run the interactive tests
        use_mpi (bool): Flag indicating whether tests are run using MPI
        num_cores (int or str): Number of cores to use (`auto` for automatic choice)
        coverage (bool): Whether to determine the test coverage
        nojit (bool): Whether to disable numba jit compilation
        early (bool): Whether to fail at the first test
        pattern (str): A pattern that determines which tests are ran

    Returns:
        int: The return code indicating success or failure
    """
    # modify current environment
    env = os.environ.copy()
    env["PYTHONPATH"] = str(PACKAGE_PATH) + ":" + env.get("PYTHONPATH", "")
    env["MPLBACKEND"] = "agg"
    if nojit:
        env["NUMBA_DISABLE_JIT"] = "1"
    else:
        env["NUMBA_WARNINGS"] = "1"
        env["NUMBA_BOUNDSCHECK"] = "1"

    if use_mpi:
        # run pytest using MPI with two cores
        args = ["mpiexec", "-n", "2"]
        if sys.platform == "darwin":
            args += ["-host", "localhost:2"]
    else:
        args = []

    # build the arguments string
    args += [
        sys.executable,
        "-m",
        "pytest",  # run pytest module
        "-c",
        "pyproject.toml",  # locate the configuration file
        "-rs",  # show summary of skipped tests
        "-rw",  # show summary of warnings raised during tests
        # "--import-mode=importlib",
    ]

    if runslow:
        args.append("--runslow")  # also run slow tests
    if runinteractive:
        args.append("--runinteractive")  # also run interactive tests
    if use_mpi:
        try:
            import numba_mpi  # @UnusedImport
        except ImportError:
            raise RuntimeError("Moduled `numba_mpi` is required to test with MPI")
        args.append("--use_mpi")  # only run tests requiring MPI multiprocessing

    # fail early if requested
    if early:
        args.append("--maxfail=1")

    # run tests using multiple cores?
    if num_cores == "auto":
        num_cores = os.cpu_count()
    else:
        num_cores = int(num_cores)
    if num_cores > 1:
        args.extend(["-n", str(num_cores), "--durations=10"])

    # run only a subset of the tests?
    if pattern is not None:
        args.extend(["-k", str(pattern)])

    # add coverage attributes?
    if coverage:
        args.extend(
            [
                "--cov-config=pyproject.toml",  # locate the configuration file
                "--cov-report",
                "html:scripts/coverage",  # create a report in html format
                f"--cov={PACKAGE}",  # specify in which package the coverage is measured
            ]
        )
        if use_mpi:
            # this is a hack to allow appending the coverage report
            args.append("--cov-append")

    # specify the package to run
    args.append("tests")

    # actually run the test
    return sp.run(args, env=env, cwd=PACKAGE_PATH).returncode


def main() -> int:
    """the main program controlling the tests

    Returns:
        int: The return code indicating success or failure
    """
    # parse the command line arguments
    parser = argparse.ArgumentParser(
        description=f"Run tests of the `{PACKAGE}` package.",
        epilog="All test categories are run if no specific categories are selected.",
    )

    # add the basic tests that need to be run
    group = parser.add_argument_group("Test categories")
    group.add_argument(
        "-s", "--style", action="store_true", default=False, help="Test code style"
    )
    group.add_argument(
        "-t", "--types", action="store_true", default=False, help="Test object types"
    )
    group.add_argument(
        "-u", "--unit", action="store_true", default=False, help="Run unit tests"
    )

    # set additional arguments
    group = parser.add_argument_group("Additional arguments")
    group.add_argument(
        "-q",
        "--quite",
        action="store_true",
        default=False,
        help="Suppress output from the script",
    )
    group.add_argument(
        "--runslow",
        action="store_true",
        default=False,
        help="Also run slow unit tests",
    )
    group.add_argument(
        "--runinteractive",
        action="store_true",
        default=False,
        help="Also run interactive unit tests",
    )
    group.add_argument(
        "--use_mpi",
        action="store_true",
        default=False,
        help="Run each unit test with MPI multiprocessing",
    )
    group.add_argument(
        "--showconfig",
        action="store_true",
        default=False,
        help="Show configuration before running tests",
    )
    group.add_argument(
        "--coverage",
        action="store_true",
        default=False,
        help="Record test coverage of unit tests",
    )
    group.add_argument(
        "--num_cores",
        metavar="CORES",
        type=str,
        default=1,
        help="Number of cores to use (`auto` for automatic choice)",
    )
    group.add_argument(
        "--nojit",
        action="store_true",
        default=False,
        help="Do not use just-in-time compilation of numba",
    )
    group.add_argument(
        "--early",
        action="store_true",
        default=False,
        help="Return at first failed test",
    )
    group.add_argument(
        "--pattern",
        metavar="PATTERN",
        type=str,
        help="Only run tests with this pattern",
    )
    group.add_argument(
        "--report",
        action="store_true",
        default=False,
        help="Write a report of the results",
    )

    # parse the command line arguments
    args = parser.parse_args()
    run_all = not (args.style or args.types or args.unit or args.showconfig)

    # show the package configuration
    if args.showconfig:
        show_config()

    # run the requested tests
    retcodes = []
    if run_all or args.style:
        retcode = test_codestyle(verbose=not args.quite)
        retcodes.append(retcode)

    if run_all or args.types:
        retcode = test_types(report=args.report, verbose=not args.quite)
        retcodes.append(retcode)

    if run_all or args.unit:
        retcode = run_unit_tests(
            runslow=args.runslow,
            runinteractive=args.runinteractive,
            use_mpi=args.use_mpi,
            coverage=args.coverage,
            num_cores=args.num_cores,
            nojit=args.nojit,
            early=args.early,
            pattern=args.pattern,
        )
        retcodes.append(retcode)

    # return the most severe code
    return _most_severe_exit_code(retcodes)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
Code for creating performance plots

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import os

os.environ["NUMBA_NUM_THREADS"] = "1"  # check single thread performance

import functools
import timeit

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from pde import ScalarField, UnitGrid
from pde.tools.misc import estimate_computation_speed
from pde.tools.output import display_progress

try:
    import cv2
except ImportError:
    print("Warning: OpenCV is not available and will thus not appear in the comparison")
    opencv_laplace = None
else:
    opencv_laplace = functools.partial(
        cv2.Laplacian, ddepth=cv2.CV_64F, borderType=cv2.BORDER_REFLECT
    )


def time_function(func, arg, repeat=3, use_out=False):
    """estimates the computation speed of a function

    Args:
        func (callable): The function to test
        arg: The single argument on which the function will be estimate
        repeat (int): How often the function is tested
        use_out (bool): Whether to supply the `out` argument to the function

    Returns:
        float: Estimated duration of calling the function a single time
    """
    if use_out:
        out = np.empty(tuple(s - 2 for s in arg.shape))
        number = int(estimate_computation_speed(func, arg, out))
        func = functools.partial(func, arg, out)
    else:
        number = int(estimate_computation_speed(func, arg))
        func = functools.partial(func, arg)
    return min(timeit.repeat(func, number=number, repeat=repeat)) / number


def get_performance_data(periodic=False):
    """obtain the data used in the performance plot

    Args:
        periodic (bool): The boundary conditions of the underlying grid

    Returns:
        dict: The durations of calculating the Laplacian on different grids
        using different methods
    """
    sizes = 2 ** np.arange(3, 13)

    statistics = {}
    for size in display_progress(sizes):
        data = {}
        grid = UnitGrid([size] * 2, periodic=periodic)
        field = ScalarField.random_normal(grid)
        field.set_ghost_cells(bc="auto_periodic_neumann")

        for backend in ["numba", "scipy"]:
            op = grid.make_operator(
                "laplace", bc="auto_periodic_neumann", backend=backend
            )
            data[backend] = time_function(op, field.data)

        op = grid.make_operator_no_bc("laplace", backend="numba")
        data["numba_no_bc"] = time_function(op, field._data_full, use_out=True)

        if opencv_laplace:
            data["opencv"] = time_function(opencv_laplace, field.data)

        statistics[int(size)] = data

    return statistics


def plot_performance(performance_data, title=None):
    """plot the performance data

    Args:
        performance_data: The data obtained from calling
            :func:`get_performance_data`.
        title (str): The title of the plot
    """
    plt.figure(figsize=[4, 3])

    PLOT_DATA = [
        {"key": "scipy", "label": "scipy", "fmt": "C0.-"},
        {"key": "opencv", "label": "opencv", "fmt": "C1.-"},
        {"key": "numba", "label": "py-pde", "fmt": "C2.-"},
        {"key": "numba_no_bc", "label": "py-pde (no BCs)", "fmt": "C2:"},
    ]

    sizes = np.array(sorted(performance_data.keys()))
    grid_sizes = sizes**2

    for plot in PLOT_DATA:
        data = np.array([performance_data[size][plot["key"]] for size in sizes])
        plt.loglog(grid_sizes, data, plot["fmt"], label=plot["label"])

    plt.xlim(grid_sizes[0], grid_sizes[-1])
    plt.xlabel("Number of grid points")
    plt.ylabel("Runtime [ms]")
    plt.legend(loc="best")

    # fix ticks of y-axis
    locmaj = mpl.ticker.LogLocator(base=10, numticks=12)
    plt.gca().xaxis.set_major_locator(locmaj)

    if title:
        plt.title(title)

    plt.tight_layout()


def main():
    """run main scripts"""
    data = get_performance_data(periodic=False)
    plot_performance(data, title="2D Laplacian (reflecting BCs)")
    plt.savefig("performance_noflux.pdf", transparent=True)
    plt.savefig("performance_noflux.png", transparent=True, dpi=200)
    plt.close()

    data = get_performance_data(periodic=True)
    plot_performance(data, title="2D Laplacian (periodic BCs)")
    plt.savefig("performance_periodic.pdf", transparent=True)
    plt.savefig("performance_periodic.png", transparent=True, dpi=200)
    plt.close()


if __name__ == "__main__":
    main()



r"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>

This module handles the boundaries of a single axis of a grid. There are
generally only two options, depending on whether the axis of the underlying
grid is defined as periodic or not. If it is periodic, the class 
:class:`~pde.grids.boundaries.axis.BoundaryPeriodic` should be used, while
non-periodic axes have more option, which are represented by
:class:`~pde.grids.boundaries.axis.BoundaryPair`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Tuple, Union

import numpy as np
from numba.extending import register_jitable

from ...tools.typing import GhostCellSetter
from ..base import GridBase, PeriodicityError
from .local import _MPIBC, BCBase, BCDataError, BoundaryData, _PeriodicBC

if TYPE_CHECKING:
    from .._mesh import GridMesh  # @UnusedImport

BoundaryPairData = Union[
    Dict[str, BoundaryData],
    BoundaryData,
    Tuple[BoundaryData, BoundaryData],
    "BoundaryAxisBase",
]


class BoundaryAxisBase:
    """base class for defining boundaries of a single axis in a grid"""

    low: BCBase
    """:class:`~pde.grids.boundaries.local.BCBase`: Boundary condition at lower end """
    high: BCBase
    """:class:`~pde.grids.boundaries.local.BCBase`: Boundary condition at upper end """

    def __init__(self, low: BCBase, high: BCBase):
        """
        Args:
            low (:class:`~pde.grids.boundaries.local.BCBase`):
                Instance describing the lower boundary
            high (:class:`~pde.grids.boundaries.local.BCBase`):
                Instance describing the upper boundary
        """
        # check data consistency
        if (
            low.grid != high.grid
            or low.axis != high.axis
            or low.rank != high.rank
            or not high.upper
            or low.upper
        ):
            raise ValueError(f"Incompatible {low!r} and {high!r}")

        # check consistency with the grid
        if not (
            low.grid.periodic[low.axis]
            == isinstance(low, _PeriodicBC)
            == isinstance(high, _PeriodicBC)
        ):
            raise PeriodicityError("Periodicity of conditions must match grid")

        self.low = low
        self.high = high

    def __repr__(self):
        return f"{self.__class__.__name__}({self.low!r}, {self.high!r})"

    def __str__(self):
        if self.low == self.high:
            return str(self.low)
        else:
            return f"({self.low}, {self.high})"

    @classmethod
    def get_help(cls) -> str:
        """Return information on how boundary conditions can be set"""
        return (
            "Boundary conditions for each side can be set using a tuple: "
            f"(lower_bc, upper_bc). {BCBase.get_help()}"
        )

    def copy(self) -> BoundaryAxisBase:
        """return a copy of itself, but with a reference to the same grid"""
        return self.__class__(self.low.copy(), self.high.copy())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.__class__ == other.__class__
            and self.low == other.low
            and self.high == other.high
        )

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.__class__ != other.__class__
            or self.low != other.low
            or self.high != other.high
        )

    def __iter__(self):
        yield self.low
        yield self.high

    def __getitem__(self, index) -> BCBase:
        """returns one of the sides"""
        if index == 0 or index is False:
            return self.low
        elif index == 1 or index is True:
            return self.high
        else:
            raise IndexError("Index must be 0/False or 1/True")

    def __setitem__(self, index, data) -> None:
        """set one of the sides"""
        # determine which side was selected
        upper = {0: False, False: False, 1: True, True: True}[index]

        # create the appropriate boundary condition
        bc = BCBase.from_data(
            self.grid,
            self.axis,
            upper=upper,
            data=data,
            rank=self[upper].rank,
        )

        # set the data
        if upper:
            self.high = bc
        else:
            self.low = bc

    @property
    def grid(self) -> GridBase:
        """:class:`~pde.grids.base.GridBase`: Underlying grid"""
        return self.low.grid

    @property
    def axis(self) -> int:
        """int: The axis along which the boundaries are defined"""
        return self.low.axis

    @property
    def periodic(self) -> bool:
        """bool: whether the axis is periodic"""
        return self.grid.periodic[self.axis]

    @property
    def rank(self) -> int:
        """int: rank of the associated boundary condition"""
        assert self.low.rank == self.high.rank
        return self.low.rank

    def get_mathematical_representation(self, field_name: str = "C") -> Tuple[str, str]:
        """return mathematical representation of the boundary condition"""
        return (
            self.low.get_mathematical_representation(field_name),
            self.high.get_mathematical_representation(field_name),
        )

    def get_data(self, idx: Tuple[int, ...]) -> Tuple[float, Dict[int, float]]:
        """sets the elements of the sparse representation of this condition

        Args:
            idx (tuple):
                The index of the point that must lie on the boundary condition

        Returns:
            float, dict: A constant value and a dictionary with indices and
            factors that can be used to calculate this virtual point
        """
        axis_coord = idx[self.axis]
        if axis_coord == -1:
            # the virtual point on the lower side
            return self.low.get_data(idx)
        elif axis_coord == self.grid.shape[self.axis]:
            # the virtual point on the upper side
            return self.high.get_data(idx)
        else:
            # the normal case of an interior point
            return 0, {axis_coord: 1}

    def set_ghost_cells(self, data_full: np.ndarray, *, args=None) -> None:
        """set the ghost cell values for all boundaries

        Args:
            data_full (:class:`~numpy.ndarray`):
                The full field data including ghost points
            args:
                Additional arguments that might be supported by special boundary
                conditions.
        """
        # send boundary information to other nodes if using MPI
        if isinstance(self.low, _MPIBC):
            self.low.send_ghost_cells(data_full, args=args)
        if isinstance(self.high, _MPIBC):
            self.high.send_ghost_cells(data_full, args=args)
        # set the actual ghost cells
        self.high.set_ghost_cells(data_full, args=args)
        self.low.set_ghost_cells(data_full, args=args)

    def make_ghost_cell_setter(self) -> GhostCellSetter:
        """return function that sets the ghost cells for this axis on a full array"""
        # get the functions that handle the data
        ghost_cell_sender_low = self.low.make_ghost_cell_sender()
        ghost_cell_sender_high = self.high.make_ghost_cell_sender()
        ghost_cell_setter_low = self.low.make_ghost_cell_setter()
        ghost_cell_setter_high = self.high.make_ghost_cell_setter()

        @register_jitable
        def ghost_cell_setter(data_full: np.ndarray, args=None) -> None:
            """helper function setting the conditions on all axes"""
            # send boundary information to other nodes if using MPI
            ghost_cell_sender_low(data_full, args=args)
            ghost_cell_sender_high(data_full, args=args)
            # set the actual ghost cells
            ghost_cell_setter_high(data_full, args=args)
            ghost_cell_setter_low(data_full, args=args)

        return ghost_cell_setter  # type: ignore


class BoundaryPair(BoundaryAxisBase):
    """represents the two boundaries of an axis along a single dimension"""

    @classmethod
    def from_data(cls, grid: GridBase, axis: int, data, rank: int = 0) -> BoundaryPair:
        """create boundary pair from some data

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid for which the boundary conditions are defined
            axis (int):
                The axis to which this boundary condition is associated
            data (str or dict):
                Data that describes the boundary pair
            rank (int):
                The tensorial rank of the field for this boundary condition

        Returns:
            :class:`~pde.grids.boundaries.axis.BoundaryPair`:
            the instance created from the data

        Throws:
            ValueError if `data` cannot be interpreted as a boundary pair
        """
        # handle the simple cases
        if isinstance(data, dict):
            if "low" in data or "high" in data:
                # separate conditions for low and high
                data_copy = data.copy()
                d_low = data_copy.pop("low")
                low = BCBase.from_data(grid, axis, upper=False, data=d_low, rank=rank)
                d_high = data_copy.pop("high")
                high = BCBase.from_data(grid, axis, upper=True, data=d_high, rank=rank)
                if data_copy:
                    raise BCDataError(f"Data items {data_copy.keys()} were not used.")
            else:
                # one condition for both sides
                low = BCBase.from_data(grid, axis, upper=False, data=data, rank=rank)
                high = BCBase.from_data(grid, axis, upper=True, data=data, rank=rank)

        elif isinstance(data, (str, BCBase)):
            # a type for both boundaries
            low = BCBase.from_data(grid, axis, upper=False, data=data, rank=rank)
            high = BCBase.from_data(grid, axis, upper=True, data=data, rank=rank)

        else:
            # the only remaining valid format is a list of conditions for the
            # lower and upper boundary
            try:
                # try obtaining the length
                data_len = len(data)
            except TypeError:
                # if len is not supported, the format must be wrong
                raise BCDataError(
                    f"Unsupported boundary format: `{data}`. " + cls.get_help()
                )
            else:
                if data_len == 2:
                    # assume that data is given for each boundary
                    low = BCBase.from_data(
                        grid, axis, upper=False, data=data[0], rank=rank
                    )
                    high = BCBase.from_data(
                        grid, axis, upper=True, data=data[1], rank=rank
                    )
                else:
                    # if the length is strange, the format must be wrong
                    raise BCDataError(
                        "Expected two conditions for the two sides of the axis, but "
                        f"got `{data}`. " + cls.get_help()
                    )

        return cls(low, high)

    def check_value_rank(self, rank: int) -> None:
        """check whether the values at the boundaries have the correct rank

        Args:
            rank (int):
                The tensorial rank of the field for this boundary condition

        Throws:
            RuntimeError: if the value does not have rank `rank`
        """
        self.low.check_value_rank(rank)
        self.high.check_value_rank(rank)


class BoundaryPeriodic(BoundaryPair):
    """represent a periodic axis"""

    def __init__(self, grid: GridBase, axis: int, flip_sign: bool = False):
        """
        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid for which the boundary conditions are defined
            axis (int):
                The axis to which this boundary condition is associated
            flip_sign (bool):
                Impose different signs on the two sides of the boundary
        """
        low = _PeriodicBC(grid=grid, axis=axis, upper=False, flip_sign=flip_sign)
        high = _PeriodicBC(grid=grid, axis=axis, upper=True, flip_sign=flip_sign)
        super().__init__(low, high)

    @property
    def flip_sign(self):
        """bool: Whether different signs are imposed on the two sides of the boundary"""
        return self.low.flip_sign

    def __repr__(self):
        res = f"{self.__class__.__name__}(grid={self.grid}, axis={self.axis}"
        if self.flip_sign:
            return res + ", flip_sign=True)"
        else:
            return res + ")"

    def __str__(self):
        if self.flip_sign:
            return '"anti-periodic"'
        else:
            return '"periodic"'

    def copy(self) -> BoundaryPeriodic:
        """return a copy of itself, but with a reference to the same grid"""
        return self.__class__(grid=self.grid, axis=self.axis, flip_sign=self.flip_sign)

    def check_value_rank(self, rank: int) -> None:
        """check whether the values at the boundaries have the correct rank

        Args:
            rank (int):
                The tensorial rank of the field for this boundary condition
        """


def get_boundary_axis(
    grid: GridBase, axis: int, data, rank: int = 0
) -> BoundaryAxisBase:
    """return object representing the boundary condition for a single axis

    Args:
        grid (:class:`~pde.grids.base.GridBase`):
            The grid for which the boundary conditions are defined
        axis (int):
            The axis to which this boundary condition is associated
        data (str or tuple or dict):
            Data describing the boundary conditions for this axis
        rank (int):
            The tensorial rank of the field for this boundary condition

    Returns:
        :class:`~pde.grids.boundaries.axis.BoundaryAxisBase`:
            Appropriate boundary condition for the axis
    """
    # handle special constructs that describe boundary conditions
    if (
        data == "natural"
        or data == "auto_periodic_neumann"
        or data == "auto_periodic_derivative"
    ):
        # automatic choice between periodic and Neumann condition
        if grid.periodic[axis]:
            data = "periodic"
        else:
            data = "derivative"

    elif data == "auto_periodic_dirichlet" or data == "auto_periodic_value":
        # automatic choice between periodic and Dirichlet condition
        if grid.periodic[axis]:
            data = "periodic"
        else:
            data = "value"

    # handle different types of data that specify boundary conditions
    if isinstance(data, BoundaryAxisBase):
        # boundary is already an the correct format
        bcs = data
    elif data == "periodic" or data == ("periodic", "periodic"):
        # initialize a periodic boundary condition
        bcs = BoundaryPeriodic(grid, axis)
    elif data == "anti-periodic" or data == ("anti-periodic", "anti-periodic"):
        # initialize a anti-periodic boundary condition
        bcs = BoundaryPeriodic(grid, axis, flip_sign=True)
    elif isinstance(data, dict) and data.get("type") == "periodic":
        # initialize a periodic boundary condition
        bcs = BoundaryPeriodic(grid, axis)
    else:
        # initialize independent boundary conditions for the two sides
        bcs = BoundaryPair.from_data(grid, axis, data, rank=rank)

    # check consistency
    if bcs.periodic != grid.periodic[axis]:
        raise PeriodicityError("Periodicity of conditions must match grid")
    return bcs
